from typing import Sequence
from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    jaxtyped,
    Array,
    DataType,
    POSE_D_O as D_O,
    NumPySampledObstacleStates,
    NumPySampledObstaclePositions,
    NumPySampledObstacleHeadings,
    NumPyObstacleStates,
    NumPyObstacleStatesForTimeStep,
    NumPyObstaclePositions,
    NumPyObstaclePositionsForTimeStep,
    NumPyObstacleOrientations,
    NumPyObstacleOrientationsForTimeStep,
)

from numtypes import D
from jaxtyping import Float

import numpy as np

ObstacleCovarianceArray = Float[Array, "T D_o D_o K"]


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class NumPySampledObstacle2dPoses(NumPySampledObstacleStates):
    """Sampled 2D poses (x, y, heading) with shape (T, POSE_D_O, K, N)."""

    _x: Float[Array, "T K N"]
    _y: Float[Array, "T K N"]
    _heading: Float[Array, "T K N"]

    @staticmethod
    def create(
        *,
        x: Float[Array, "T K N"],
        y: Float[Array, "T K N"],
        heading: Float[Array, "T K N"],
    ) -> "NumPySampledObstacle2dPoses":
        return NumPySampledObstacle2dPoses(_x=x, _y=y, _heading=heading)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_o K N"]:
        return self.array

    def x(self) -> Float[Array, "T K N"]:
        return self._x

    def y(self) -> Float[Array, "T K N"]:
        return self._y

    def heading(self) -> Float[Array, "T K N"]:
        return self._heading

    def positions(self) -> NumPySampledObstaclePositions:
        return NumPySampledObstaclePositions.create(x=self._x, y=self._y)

    def headings(self) -> NumPySampledObstacleHeadings:
        return NumPySampledObstacleHeadings.create(heading=self._heading)

    def at(self, *, time_step: int, sample: int) -> "NumPyObstacle2dPosesForTimeStep":
        return NumPyObstacle2dPosesForTimeStep.create(
            x=self._x[time_step, :, sample],
            y=self._y[time_step, :, sample],
            heading=self._heading[time_step, :, sample],
        )

    @property
    def horizon(self) -> int:
        return self._x.shape[0]

    @property
    def dimension(self) -> int:
        return D_O

    @property
    def count(self) -> int:
        return self._x.shape[1]

    @property
    def sample_count(self) -> int:
        return self._x.shape[2]

    @property
    def array(self) -> Float[Array, "T D_o K N"]:
        return self._array

    @cached_property
    def _array(self) -> Float[Array, "T D_o K N"]:
        return np.stack([self._x, self._y, self._heading], axis=1)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class NumPyObstacle2dPositions(NumPyObstaclePositions):
    """2D positions (x, y) with shape (T, 2, K)."""

    _x: Float[Array, "T K"]
    _y: Float[Array, "T K"]

    @staticmethod
    def create(
        *, x: Float[Array, "T K"], y: Float[Array, "T K"]
    ) -> "NumPyObstacle2dPositions":
        return NumPyObstacle2dPositions(_x=x, _y=y)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T 2 K"]:
        return self.array

    @property
    def horizon(self) -> int:
        return self._x.shape[0]

    @property
    def dimension(self) -> D[2]:
        return 2

    @property
    def count(self) -> int:
        return self._x.shape[1]

    @property
    def array(self) -> Float[Array, "T 2 K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[Array, "T 2 K"]:
        return np.stack([self._x, self._y], axis=1)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class NumPyObstacle2dPositionsForTimeStep(NumPyObstaclePositionsForTimeStep):
    """2D positions (x, y) for a single time step with shape (2, K)."""

    _x: Float[Array, " K"]
    _y: Float[Array, " K"]

    @staticmethod
    def create(
        *, x: Float[Array, " K"], y: Float[Array, " K"]
    ) -> "NumPyObstacle2dPositionsForTimeStep":
        return NumPyObstacle2dPositionsForTimeStep(_x=x, _y=y)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "2 K"]:
        return self.array

    @property
    def dimension(self) -> D[2]:
        return 2

    @property
    def count(self) -> int:
        return self._x.shape[0]

    @property
    def array(self) -> Float[Array, "2 K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[Array, "2 K"]:
        return np.stack([self._x, self._y], axis=0)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class NumPyObstacleHeadings(NumPyObstacleOrientations):
    """Obstacle headings with shape (T, 1, K)."""

    _heading: Float[Array, "T K"]

    @staticmethod
    def create(*, heading: Float[Array, "T K"]) -> "NumPyObstacleHeadings":
        return NumPyObstacleHeadings(_heading=heading)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T 1 K"]:
        return self.array

    @property
    def horizon(self) -> int:
        return self._heading.shape[0]

    @property
    def dimension(self) -> D[1]:
        return 1

    @property
    def count(self) -> int:
        return self._heading.shape[1]

    @property
    def array(self) -> Float[Array, "T 1 K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[Array, "T 1 K"]:
        return self._heading[:, np.newaxis, :]


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class NumPyObstacleHeadingsForTimeStep(NumPyObstacleOrientationsForTimeStep):
    """Obstacle headings for a single time step."""

    _heading: Float[Array, " K"]

    @staticmethod
    def create(*, heading: Float[Array, " K"]) -> "NumPyObstacleHeadingsForTimeStep":
        return NumPyObstacleHeadingsForTimeStep(_heading=heading)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "1 K"]:
        return self.array

    @property
    def dimension(self) -> D[1]:
        return 1

    @property
    def count(self) -> int:
        return self._heading.shape[0]

    @property
    def array(self) -> Float[Array, "1 K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[Array, "1 K"]:
        return self._heading[np.newaxis, :]


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class NumPyObstacle2dPoses(
    NumPyObstacleStates[
        NumPySampledObstacle2dPoses,
        "NumPyObstacle2dPosesForTimeStep",
    ]
):
    """2D poses (x, y, heading) with shape (T, POSE_D_O, K)."""

    _x: Float[Array, "T K"]
    _y: Float[Array, "T K"]
    _heading: Float[Array, "T K"]
    _covariance: ObstacleCovarianceArray | None

    @staticmethod
    def empty(*, horizon: int, obstacle_count: int = 0) -> "NumPyObstacle2dPoses":
        """Creates obstacle states for zero obstacles over the given time horizon."""
        empty = np.full((horizon, obstacle_count), fill_value=np.nan)

        return NumPyObstacle2dPoses.create(x=empty, y=empty, heading=empty)

    @staticmethod
    def wrap(array: Float[Array, "T D_o K"]) -> "NumPyObstacle2dPoses":
        return NumPyObstacle2dPoses.create(
            x=array[:, 0, :], y=array[:, 1, :], heading=array[:, 2, :]
        )

    @staticmethod
    def sampled(  # type: ignore
        *,
        x: Float[Array, "T K N"],
        y: Float[Array, "T K N"],
        heading: Float[Array, "T K N"],
    ) -> NumPySampledObstacle2dPoses:
        return NumPySampledObstacle2dPoses.create(x=x, y=y, heading=heading)

    @staticmethod
    def create(
        *,
        x: Float[Array, "T K"],
        y: Float[Array, "T K"],
        heading: Float[Array, "T K"],
        covariance: ObstacleCovarianceArray | None = None,
    ) -> "NumPyObstacle2dPoses":
        return NumPyObstacle2dPoses(
            _x=x, _y=y, _heading=heading, _covariance=covariance
        )

    @staticmethod
    def of_states(
        obstacle_states: Sequence["NumPyObstacle2dPosesForTimeStep"],
    ) -> "NumPyObstacle2dPoses":
        assert len(obstacle_states) > 0, "Obstacle states sequence must not be empty."

        T = len(obstacle_states)
        K = max(states.count for states in obstacle_states)

        x = np.full((T, K), np.nan)
        y = np.full((T, K), np.nan)
        heading = np.full((T, K), np.nan)

        for t, states in enumerate(obstacle_states):
            n = states.count
            x[t, :n] = states.x()
            y[t, :n] = states.y()
            heading[t, :n] = states.heading()

        return NumPyObstacle2dPoses.create(x=x, y=y, heading=heading)

    @staticmethod
    def for_time_step(
        *, x: Float[Array, " K"], y: Float[Array, " K"], heading: Float[Array, " K"]
    ) -> "NumPyObstacle2dPosesForTimeStep":
        return NumPyObstacle2dPosesForTimeStep.create(x=x, y=y, heading=heading)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_o K"]:
        return self.array

    def x(self) -> Float[Array, "T K"]:
        return self._x

    def y(self) -> Float[Array, "T K"]:
        return self._y

    def heading(self) -> Float[Array, "T K"]:
        return self._heading

    def positions(self) -> NumPyObstacle2dPositions:
        return NumPyObstacle2dPositions.create(x=self._x, y=self._y)

    def headings(self) -> NumPyObstacleHeadings:
        return NumPyObstacleHeadings.create(heading=self._heading)

    def covariance(self) -> ObstacleCovarianceArray | None:
        return self._covariance

    def single(self) -> NumPySampledObstacle2dPoses:
        return NumPySampledObstacle2dPoses.create(
            x=self._x[..., np.newaxis],
            y=self._y[..., np.newaxis],
            heading=self._heading[..., np.newaxis],
        )

    def last(self) -> "NumPyObstacle2dPosesForTimeStep":
        return self.at(time_step=self.horizon - 1)

    def at(self, *, time_step: int) -> "NumPyObstacle2dPosesForTimeStep":
        return NumPyObstacle2dPosesForTimeStep.create(
            x=self._x[time_step],
            y=self._y[time_step],
            heading=self._heading[time_step],
        )

    @property
    def horizon(self) -> int:
        return self._x.shape[0]

    @property
    def dimension(self) -> int:
        return D_O

    @property
    def count(self) -> int:
        return self._x.shape[1]

    @property
    def array(self) -> Float[Array, "T D_o K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[Array, "T D_o K"]:
        return np.stack([self._x, self._y, self._heading], axis=1)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class NumPyObstacle2dPosesForTimeStep(
    NumPyObstacleStatesForTimeStep[NumPyObstacle2dPoses]
):
    """2D poses (x, y, heading) for a single time step."""

    _x: Float[Array, " K"]
    _y: Float[Array, " K"]
    _heading: Float[Array, " K"]

    @staticmethod
    def wrap(array: Float[Array, f"{D_O} K"]) -> "NumPyObstacle2dPosesForTimeStep":
        return NumPyObstacle2dPosesForTimeStep.create(
            x=array[0, :], y=array[1, :], heading=array[2, :]
        )

    @staticmethod
    def create(
        *,
        x: Float[Array, " K"],
        y: Float[Array, " K"],
        heading: Float[Array, " K"],
    ) -> "NumPyObstacle2dPosesForTimeStep":
        return NumPyObstacle2dPosesForTimeStep(_x=x, _y=y, _heading=heading)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "D_o K"]:
        return self.array

    def x(self) -> Float[Array, " K"]:
        return self._x

    def y(self) -> Float[Array, " K"]:
        return self._y

    def heading(self) -> Float[Array, " K"]:
        return self._heading

    def positions(self) -> NumPyObstacle2dPositionsForTimeStep:
        return NumPyObstacle2dPositionsForTimeStep.create(x=self._x, y=self._y)

    def headings(self) -> NumPyObstacleHeadingsForTimeStep:
        return NumPyObstacleHeadingsForTimeStep.create(heading=self._heading)

    def replicate(self, *, horizon: int) -> NumPyObstacle2dPoses:
        return NumPyObstacle2dPoses.create(
            x=np.tile(self._x[np.newaxis, :], (horizon, 1)),
            y=np.tile(self._y[np.newaxis, :], (horizon, 1)),
            heading=np.tile(self._heading[np.newaxis, :], (horizon, 1)),
        )

    @property
    def dimension(self) -> int:
        return D_O

    @property
    def count(self) -> int:
        return self._x.shape[0]

    @property
    def array(self) -> Float[Array, "D_o K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[Array, "D_o K"]:
        return np.stack([self._x, self._y, self._heading], axis=0)



from typing import Any
from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    jaxtyped,
    Array,
    StateSequence,
    ContouringCost,
    LagCost,
    SimulationData,
    Metric,
)
from faran.collectors import access

from jaxtyping import Float

import numpy as np


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class MpccErrorMetricResult:
    """Results of the MPCC error metric, including contouring and lag errors.

    Attributes:
        contouring: Contouring error at each time step.
        lag: Lag error at each time step.
    """

    contouring: Float[Array, " T"]
    lag: Float[Array, " T"]

    @cached_property
    def max_contouring(self) -> float:
        """Returns the maximum absolute contouring error detected."""
        return float(np.abs(self.contouring).max())

    @cached_property
    def max_lag(self) -> float:
        """Returns the maximum absolute lag error detected."""
        return float(np.abs(self.lag).max())


@dataclass(kw_only=True, frozen=True)
class MpccErrorMetric[StateBatchT](Metric[MpccErrorMetricResult]):
    """Metric evaluating contouring and lag errors in MPCC tracking."""

    contouring: ContouringCost[Any, StateBatchT]
    lag: LagCost[Any, StateBatchT]

    @staticmethod
    def create(
        *,
        contouring: ContouringCost[Any, StateBatchT],
        lag: LagCost[Any, StateBatchT],
    ) -> "MpccErrorMetric[StateBatchT]":
        return MpccErrorMetric(contouring=contouring, lag=lag)

    def compute(self, data: SimulationData) -> MpccErrorMetricResult:
        states = data(access.states.assume(StateSequence[StateBatchT]).require())
        state_batch = states.batched()

        contouring_error = self.contouring.error(states=state_batch)
        lag_error = self.lag.error(states=state_batch)

        return MpccErrorMetricResult(
            contouring=np.asarray(contouring_error)[..., 0],
            lag=np.asarray(lag_error)[..., 0],
        )

    @property
    def name(self) -> str:
        return "mpcc-error"



from typing import Any
from dataclasses import dataclass

from faran.types import (
    jaxtyped,
    Array,
    DataType,
    StateBatch,
    ControlInputBatch,
    CostFunction,
    ContouringCost,
    LagCost,
    Error,
    Trajectory,
    NumPyControlInputBatch,
    NumPyCosts,
    NumPyPathParameters,
    NumPyReferencePoints,
    NumPyPositionExtractor,
    NumPyPathParameterExtractor,
    NumPyPathVelocityExtractor,
)
from faran.states import NumPySimpleCosts

from jaxtyping import Float

import numpy as np


@jaxtyped
@dataclass(frozen=True)
class NumPyError(Error):
    """Contouring or lag error between the state batch and reference trajectory."""

    array: Float[Array, "T M"]

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T M"]:
        return self.array


@dataclass(kw_only=True, frozen=True)
class NumPyContouringCost[StateBatchT](
    ContouringCost[ControlInputBatch, StateBatchT, NumPyError],
    CostFunction[ControlInputBatch, StateBatchT, NumPyCosts],
):
    """Penalizes lateral deviation from a reference trajectory."""

    reference: Trajectory[NumPyPathParameters, NumPyReferencePoints]
    path_parameter_extractor: NumPyPathParameterExtractor[StateBatchT]
    position_extractor: NumPyPositionExtractor[StateBatchT]
    weight: float

    @staticmethod
    def create[S](
        *,
        reference: Trajectory[NumPyPathParameters, NumPyReferencePoints],
        path_parameter_extractor: NumPyPathParameterExtractor[S],
        position_extractor: NumPyPositionExtractor[S],
        weight: float,
    ) -> "NumPyContouringCost[S]":
        """Creates a contouring cost implemented with NumPy.

        Args:
            reference: The reference trajectory to follow.
            path_parameter_extractor: Extracts the path parameters from a state batch.
            position_extractor: Extracts the (x, y) positions from a state batch.
            weight: The weight of the contouring cost.
        """
        return NumPyContouringCost(
            reference=reference,
            path_parameter_extractor=path_parameter_extractor,
            position_extractor=position_extractor,
            weight=weight,
        )

    def __call__(self, *, inputs: ControlInputBatch, states: StateBatchT) -> NumPyCosts:
        error = self.error(states=states)
        return NumPySimpleCosts(self.weight * error.array**2)

    def error(self, *, states: StateBatchT) -> NumPyError:
        ref_points = self.reference.query(self.path_parameter_extractor(states))
        heading = ref_points.heading()
        positions = self.position_extractor(states)

        return NumPyError(
            np.sin(heading) * (positions.x() - ref_points.x())
            - np.cos(heading) * (positions.y() - ref_points.y())
        )


@dataclass(kw_only=True, frozen=True)
class NumPyLagCost[StateBatchT](
    LagCost[ControlInputBatch, StateBatchT, NumPyError],
    CostFunction[ControlInputBatch, StateBatchT, NumPyCosts],
):
    """Penalizes longitudinal deviation from a reference trajectory."""

    reference: Trajectory[NumPyPathParameters, NumPyReferencePoints]
    path_parameter_extractor: NumPyPathParameterExtractor[StateBatchT]
    position_extractor: NumPyPositionExtractor[StateBatchT]
    weight: float

    @staticmethod
    def create[S](
        *,
        reference: Trajectory[NumPyPathParameters, NumPyReferencePoints],
        path_parameter_extractor: NumPyPathParameterExtractor[S],
        position_extractor: NumPyPositionExtractor[S],
        weight: float,
    ) -> "NumPyLagCost[S]":
        """Creates a lag cost implemented with NumPy.

        Args:
            reference: The reference trajectory to follow.
            path_parameter_extractor: Extracts the path parameters from a state batch.
            position_extractor: Extracts the (x, y) positions from a state batch.
            weight: The weight of the lag cost.
        """
        return NumPyLagCost(
            reference=reference,
            path_parameter_extractor=path_parameter_extractor,
            position_extractor=position_extractor,
            weight=weight,
        )

    def __call__(self, *, inputs: ControlInputBatch, states: StateBatchT) -> NumPyCosts:
        error = self.error(states=states)
        return NumPySimpleCosts(self.weight * error.array**2)

    def error(self, *, states: StateBatchT) -> NumPyError:
        ref_points = self.reference.query(self.path_parameter_extractor(states))
        heading = ref_points.heading()
        positions = self.position_extractor(states)

        return NumPyError(
            -np.cos(heading) * (positions.x() - ref_points.x())
            - np.sin(heading) * (positions.y() - ref_points.y())
        )


@dataclass(kw_only=True, frozen=True)
class NumPyProgressCost[InputBatchT](CostFunction[InputBatchT, StateBatch, NumPyCosts]):
    """Rewards forward progress along a reference trajectory."""

    path_velocity_extractor: NumPyPathVelocityExtractor[InputBatchT]
    time_step_size: float
    weight: float

    @staticmethod
    def create[I](
        *,
        path_velocity_extractor: NumPyPathVelocityExtractor[I],
        time_step_size: float,
        weight: float,
    ) -> "NumPyProgressCost[I]":
        """Creates a progress cost implemented with NumPy.

        Args:
            path_velocity_extractor: Extracts the path velocities from a control input batch.
            time_step_size: The time step size between states.
            weight: The weight of the progress cost.
        """
        return NumPyProgressCost(
            path_velocity_extractor=path_velocity_extractor,
            time_step_size=time_step_size,
            weight=weight,
        )

    def __call__(self, *, inputs: InputBatchT, states: StateBatch) -> NumPyCosts:
        path_velocities = self.path_velocity_extractor(inputs)

        return NumPySimpleCosts(-self.weight * path_velocities * self.time_step_size)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class NumPyControlSmoothingCost(
    CostFunction[NumPyControlInputBatch, StateBatch, NumPyCosts]
):
    """Penalizes abrupt changes in control inputs between consecutive time steps."""

    weights: Float[Array, " D_u"]

    @staticmethod
    def create(
        *,
        weights: Float[Array, " D_u"],
    ) -> "NumPyControlSmoothingCost":
        """Creates a control smoothing cost implemented with NumPy.

        Args:
            weights: The weights for each control input dimension.
        """
        return NumPyControlSmoothingCost(weights=weights)

    def __call__(
        self, *, inputs: NumPyControlInputBatch, states: StateBatch
    ) -> NumPyCosts:
        diffs = np.diff(inputs.array, axis=0, prepend=inputs.array[0:1, :, :])
        squared_diffs = diffs**2
        weighted_squared_diffs = squared_diffs * self.weights[np.newaxis, :, np.newaxis]
        cost_per_time_step = np.sum(weighted_squared_diffs, axis=1)

        return NumPySimpleCosts(cost_per_time_step)


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class NumPyControlEffortCost(CostFunction[NumPyControlInputBatch, Any, NumPyCosts]):
    """Penalizes large control input magnitudes."""

    weights: Float[Array, " D_u"]

    @staticmethod
    def create(*, weights: Float[Array, " D_u"]) -> "NumPyControlEffortCost":
        """Creates a control effort cost implemented with NumPy.

        Args:
            weights: The weights for each control input dimension.
        """
        return NumPyControlEffortCost(weights=weights)

    def __call__(self, *, inputs: NumPyControlInputBatch, states: Any) -> NumPyCosts:
        return NumPySimpleCosts(np.einsum("u,tum->tm", self.weights, inputs.array**2))



from .basic import (
    NumPyAugmentedState as NumPyAugmentedState,
    NumPyAugmentedStateSequence as NumPyAugmentedStateSequence,
    NumPyAugmentedStateBatch as NumPyAugmentedStateBatch,
    NumPyAugmentedControlInputSequence as NumPyAugmentedControlInputSequence,
    NumPyAugmentedControlInputBatch as NumPyAugmentedControlInputBatch,
)
from .accelerated import (
    JaxAugmentedState as JaxAugmentedState,
    JaxAugmentedStateSequence as JaxAugmentedStateSequence,
    JaxAugmentedStateBatch as JaxAugmentedStateBatch,
    JaxAugmentedControlInputSequence as JaxAugmentedControlInputSequence,
    JaxAugmentedControlInputBatch as JaxAugmentedControlInputBatch,
)
from .common import (
    AugmentedModel as AugmentedModel,
    AugmentedSampler as AugmentedSampler,
    extract as extract,
)
from .mppi import (
    AugmentedMppi as AugmentedMppi,
    NumPyAugmentedMppi as NumPyAugmentedMppi,
    JaxAugmentedMppi as JaxAugmentedMppi,
)



from typing import Protocol, Self, Any
from dataclasses import dataclass

from faran.types.array import Array, DataType

from jaxtyping import Float


class State(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, " D_x"]:
        """Returns the state as a NumPy array."""
        ...

    @property
    def dimension(self) -> int:
        """Returns the dimension of the state."""
        ...


class StateSequence[StateBatchT = Any](Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_x"]:
        """Returns the state sequence as a NumPy array."""
        ...

    def batched(self) -> StateBatchT:
        """Returns the state sequence as a batch with a single rollout."""
        ...

    @property
    def horizon(self) -> int:
        """Time horizon of the state sequence."""
        ...

    @property
    def dimension(self) -> int:
        """State dimension."""
        ...


class StateBatch(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_x M"]:
        """Returns the states as a NumPy array."""
        ...

    @property
    def horizon(self) -> int:
        """Time horizon of the state batch."""
        ...

    @property
    def dimension(self) -> int:
        """State dimension."""
        ...

    @property
    def rollout_count(self) -> int:
        """Number of rollouts in the batch."""
        ...


class ControlInputSequence(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_u"]:
        """Returns the control input sequence as a NumPy array."""
        ...

    @property
    def horizon(self) -> int:
        """Time horizon of the control input sequence."""
        ...

    @property
    def dimension(self) -> int:
        """Control input dimension."""
        ...


class ControlInputBatch(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_u M"]:
        """Returns the control inputs as a NumPy array."""
        ...

    @property
    def horizon(self) -> int:
        """Time horizon of the control input batch."""
        ...

    @property
    def dimension(self) -> int:
        """Control input dimension."""
        ...

    @property
    def rollout_count(self) -> int:
        """Number of rollouts in the batch."""
        ...


class DynamicalModel[StateT, StateSequenceT, StateBatchT, InputSequenceT, InputBatchT](
    Protocol
):
    def simulate(self, inputs: InputBatchT, initial_state: StateT) -> StateBatchT:
        """Simulates the dynamical model over the given control inputs starting from the
        provided initial state."""
        ...

    def step(self, inputs: InputSequenceT, state: StateT) -> StateT:
        """Simulates a single time step of the dynamical model given the control input and current
        state."""
        ...

    def forward(self, inputs: InputSequenceT, state: StateT) -> StateSequenceT:
        """Simulates the dynamical model over the given control input sequence starting from the
        provided initial state."""
        ...

    @property
    def time_step_size(self) -> float:
        """Time step size of the dynamical model."""
        ...


class Costs(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T M"]:
        """Returns the costs as a NumPy array."""
        ...

    def zero(self) -> Self:
        """Returns zero costs similar to this one."""
        ...

    @property
    def horizon(self) -> int:
        """Time horizon of the costs."""
        ...

    @property
    def rollout_count(self) -> int:
        """Number of rollouts the costs correspond to."""
        ...


class CostFunction[InputBatchT, StateBatchT, CostsT](Protocol):
    def __call__(self, *, inputs: InputBatchT, states: StateBatchT) -> CostsT:
        """Computes the cost for each time step and rollout."""
        ...


class Sampler[InputSequenceT, InputBatchT](Protocol):
    def sample(self, *, around: InputSequenceT) -> InputBatchT:
        """Samples a batch of control input sequences around the given nominal input."""
        ...

    @property
    def rollout_count(self) -> int:
        """Number of rollouts the sampler generates."""
        ...


class Weights(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, " M"]:
        """Returns the weights as a NumPy array."""
        ...

    @property
    def rollout_count(self) -> int:
        """Number of rollouts the weights correspond to."""
        ...


@dataclass(kw_only=True, frozen=True)
class DebugData[WeightsT]:
    trajectory_weights: WeightsT


@dataclass(frozen=True)
class Control[InputSequenceT, WeightsT]:
    optimal: InputSequenceT
    nominal: InputSequenceT
    debug: DebugData[WeightsT]


class Mppi[StateT, InputSequenceT, WeightsT = Any](Protocol):
    def step(
        self,
        *,
        temperature: float,
        nominal_input: InputSequenceT,
        initial_state: StateT,
    ) -> Control[InputSequenceT, WeightsT]:
        """Runs one iteration of the MPPI algorithm to compute the next optimal and nominal
        control sequences.
        """
        ...


class UpdateFunction[InputSequenceT](Protocol):
    def __call__(
        self, *, nominal_input: InputSequenceT, optimal_input: InputSequenceT
    ) -> InputSequenceT:
        """Updates the nominal control input sequence based on the optimal control input
        sequence.
        """
        ...


class PaddingFunction[NominalT, PaddingT](Protocol):
    def __call__(self, *, nominal_input: NominalT, padding_size: int) -> PaddingT:
        """Generates padding values for the shifted nominal control input sequence."""
        ...


class FilterFunction[InputSequenceT](Protocol):
    def __call__(self, *, optimal_input: InputSequenceT) -> InputSequenceT:
        """Filters the optimal control input after it is computed."""
        ...



from .factory import metrics as metrics
from .registry import MetricRegistry as MetricRegistry
from .collision import (
    CollisionMetricResult as CollisionMetricResult,
    CollisionMetric as CollisionMetric,
)
from .error import (
    MpccErrorMetricResult as MpccErrorMetricResult,
    MpccErrorMetric as MpccErrorMetric,
)
from .task import (
    TaskCompletionMetricResult as TaskCompletionMetricResult,
    TaskCompletionMetric as TaskCompletionMetric,
)
from .constraint import (
    ConstraintViolationMetricResult as ConstraintViolationMetricResult,
    ConstraintViolationMetric as ConstraintViolationMetric,
)
from .comfort import (
    ComfortMetricResult as ComfortMetricResult,
    ComfortMetric as ComfortMetric,
)



from typing import Self
from dataclasses import dataclass

from faran.types import Array, jaxtyped, NumPyObstacleSimulator
from faran.obstacles.basic import NumPyObstacle2dPosesForTimeStep

from jaxtyping import Float

import numpy as np


@jaxtyped
@dataclass(frozen=True)
class NumPyStaticObstacleSimulator(
    NumPyObstacleSimulator[NumPyObstacle2dPosesForTimeStep]
):
    """Simulates stationary obstacles by replicating fixed positions over the horizon."""

    positions: Float[Array, "K 2"]
    headings: Float[Array, " K"]

    @staticmethod
    def empty() -> "NumPyStaticObstacleSimulator":
        positions = np.empty((0, 2))

        return NumPyStaticObstacleSimulator.create(positions=positions)

    @staticmethod
    def create(
        *, positions: Float[Array, "K 2"], headings: Float[Array, " K"] | None = None
    ) -> "NumPyStaticObstacleSimulator":
        count = positions.shape[0]
        headings = headings if headings is not None else np.zeros(shape=(count,))

        return NumPyStaticObstacleSimulator(positions=positions, headings=headings)

    def with_time_step_size(self, time_step_size: float) -> Self:
        # Time step does not matter.
        return self

    def step(self) -> NumPyObstacle2dPosesForTimeStep:
        return NumPyObstacle2dPosesForTimeStep.create(
            x=self.positions[:, 0],
            y=self.positions[:, 1],
            heading=self.headings,
        )

    @property
    def obstacle_count(self) -> int:
        return self.positions.shape[0]



from typing import Protocol, Final

from faran.types.array import Array, DataType

from jaxtyping import Float
from numtypes import D

BICYCLE_D_X: Final = 4
BICYCLE_D_U: Final = 2
BICYCLE_D_O: Final = 4

type BicycleD_x = D[4]
"""State dimension of the bicycle model, consisting of (x position, y position, heading, speed)."""

type BicycleD_u = D[2]
"""Control input dimension of the bicycle model, consisting of (acceleration, steering angle)."""

type BicycleD_o = D[4]
"""Obstacle state dimension of the bicycle model, consisting of (x position, y position, heading, speed)."""


class BicycleState(Protocol):
    def __array__(
        self, dtype: DataType | None = None
    ) -> Float[Array, f"{BICYCLE_D_X}"]:
        """Returns the state as a NumPy array."""
        ...

    @property
    def x(self) -> float:
        """X position of the agent."""
        ...

    @property
    def y(self) -> float:
        """Y position of the agent."""
        ...

    @property
    def heading(self) -> float:
        """Orientation of the agent."""
        ...

    @property
    def speed(self) -> float:
        """Velocity of the agent."""
        ...

    @property
    def dimension(self) -> BicycleD_x:
        """State dimension."""
        ...


class BicycleStateSequence(Protocol):
    def step(self, index: int) -> BicycleState:
        """Returns the state at the given time step index."""
        ...


class BicycleStateBatch(Protocol):
    def __array__(
        self, dtype: DataType | None = None
    ) -> Float[Array, f"T {BICYCLE_D_X} M"]:
        """Returns the states as a NumPy array."""
        ...

    def heading(self) -> Float[Array, "T M"]:
        """Returns the headings (orientations) of the states in the batch."""
        ...

    def speed(self) -> Float[Array, "T M"]:
        """Returns the speeds of the states in the batch."""
        ...

    def rollout(self, index: int) -> BicycleStateSequence:
        """Returns a single rollout from the batch as a state sequence."""
        ...

    @property
    def positions(self) -> "BicyclePositions":
        """Returns the positions of the states in the batch."""
        ...


class BicyclePositions(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T 2 M"]:
        """Returns the positions as a NumPy array."""
        ...

    def x(self) -> Float[Array, "T M"]:
        """Returns the x positions."""
        ...

    def y(self) -> Float[Array, "T M"]:
        """Returns the y positions."""
        ...


class BicycleControlInputSequence(Protocol):
    def __array__(
        self, dtype: DataType | None = None
    ) -> Float[Array, f"T {BICYCLE_D_U}"]:
        """Returns the control input sequence as a NumPy array."""
        ...

    @property
    def horizon(self) -> int:
        """Time horizon of the control input sequence."""
        ...

    @property
    def dimension(self) -> BicycleD_u:
        """Control input dimension."""
        ...


class BicycleControlInputBatch(Protocol):
    def __array__(
        self, dtype: DataType | None = None
    ) -> Float[Array, f"T {BICYCLE_D_U} M"]:
        """Returns the control inputs as a NumPy array."""
        ...

    @property
    def rollout_count(self) -> int:
        """Number of rollouts in the batch."""
        ...

    @property
    def horizon(self) -> int:
        """Time horizon of the control inputs."""
        ...



import warnings
from typing import Sequence
from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    jaxtyped,
    Array,
    Trajectory,
    NumPyPathParameters,
    NumPyReferencePoints,
    NumPyPositions,
    NumPyLateralPositions,
    NumPyLongitudinalPositions,
    NumPyNormals,
)

from jaxtyping import Float
from scipy.interpolate import CubicSpline
from scipy.optimize import newton
from scipy.spatial import cKDTree  # type: ignore

import numpy as np


type PointArray = Float[Array, "N 2"]
type PointsDescription = PointArray | Sequence[tuple[float, float]]


class standardize:
    @staticmethod
    def points(points: PointsDescription) -> PointArray:
        if isinstance(points, Sequence):
            return np.asarray(points, dtype=np.float64)

        return points


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class NumPyWaypointsTrajectory(
    Trajectory[
        NumPyPathParameters,
        NumPyReferencePoints,
        NumPyPositions,
        NumPyLateralPositions,
        NumPyLongitudinalPositions,
    ]
):
    """NumPy cubic spline trajectory through a set of waypoints."""

    reference_points: Float[Array, " N_ref"]
    spline_x: CubicSpline
    spline_y: CubicSpline

    kd_tree_samples: int
    refining_iterations: int

    _path_length: float
    _natural_length: float

    @staticmethod
    def create(
        *,
        points: PointsDescription,
        path_length: float | None = None,
        coarse_search_samples: int = 200,
        refining_iterations: int = 3,
    ) -> "NumPyWaypointsTrajectory":
        """Creates a waypoints trajectory from a set of 2D points.

        Args:
            points: Array of shape (N, 2) containing N waypoints with (x, y) coordinates.
            path_length: Total length of the trajectory. If omitted, it will be set to the natural length
                of the path defined by the waypoints.
            coarse_search_samples: Number of samples to use for the initial coarse nearest-point search.
            refining_iterations: Number of iterations for local refinement of closest points.

        Note:
            This uses a KD-tree for efficient nearest-point queries under the hood. The `coarse_search_samples`
            parameter controls the number of samples along the spline used to build the KD-tree for the initial guess
            of the closest point. The `refining_iterations` parameter controls how many iterations of Newton's method
            are used to refine the closest point after the initial guess from the KD-tree.
        """
        points = standardize.points(points)
        x, y = points.T

        path_parameters = compute_path_parameters(x=x, y=y)
        natural_length = path_parameters[-1]
        path_length = float(natural_length) if path_length is None else path_length
        normalized_lengths = path_parameters * (path_length / natural_length)

        spline_x = CubicSpline(normalized_lengths, x, bc_type="natural")
        spline_y = CubicSpline(normalized_lengths, y, bc_type="natural")

        return NumPyWaypointsTrajectory(
            reference_points=normalized_lengths,
            spline_x=spline_x,
            spline_y=spline_y,
            kd_tree_samples=coarse_search_samples,
            refining_iterations=refining_iterations,
            _path_length=path_length,
            _natural_length=natural_length,
        )

    def query(self, parameters: NumPyPathParameters) -> NumPyReferencePoints:
        s = np.asarray(parameters)

        assert np.all((0.0 <= s) & (s <= self.path_length)), (
            f"Path parameters out of bounds. Got: {s}"
        )

        x = self.spline_x(s).astype(np.float64, copy=False)
        y = self.spline_y(s).astype(np.float64, copy=False)

        dx_ds = self.spline_x(s, 1)
        dy_ds = self.spline_y(s, 1)
        heading = np.arctan2(dy_ds, dx_ds).astype(np.float64, copy=False)

        return NumPyReferencePoints.create(x=x, y=y, heading=heading)

    def longitudinal(self, positions: NumPyPositions) -> NumPyLongitudinalPositions:
        result = self._closest_arc_lengths(positions)
        return NumPyLongitudinalPositions.create(result)

    def lateral(self, positions: NumPyPositions) -> NumPyLateralPositions:
        s = self._closest_arc_lengths(positions)

        closest_x = self.spline_x(s)
        closest_y = self.spline_y(s)
        dx_ds = self.spline_x(s, 1)
        dy_ds = self.spline_y(s, 1)
        tangent_norm = np.sqrt(dx_ds**2 + dy_ds**2)

        diff_x = positions.x() - closest_x
        diff_y = positions.y() - closest_y

        lateral = (diff_x * dy_ds - diff_y * dx_ds) / tangent_norm

        return NumPyLateralPositions.create(np.asarray(lateral, dtype=np.float64))

    def normal(self, parameters: NumPyPathParameters) -> NumPyNormals:
        s = np.asarray(parameters)

        dx_ds = self.spline_x(s, 1)
        dy_ds = self.spline_y(s, 1)
        norm = np.sqrt(dx_ds**2 + dy_ds**2)

        x = dy_ds / norm
        y = -dx_ds / norm

        return NumPyNormals.create(x=x, y=y)

    @property
    def end(self) -> tuple[float, float]:
        return (
            float(self.spline_x(self.path_length)),
            float(self.spline_y(self.path_length)),
        )

    @property
    def path_length(self) -> float:
        return self._path_length

    @property
    def natural_length(self) -> float:
        return self._natural_length

    def _closest_arc_lengths(self, positions: NumPyPositions) -> Float[Array, "T M"]:
        return self._refine(self._nearest_samples_to(positions), positions)

    def _nearest_samples_to(self, positions: NumPyPositions) -> Float[Array, "T M"]:
        # Finds the nearest samples along the spline for given positions.
        T, M = positions.horizon, positions.rollout_count
        arc_lengths, tree = self._guess_samples

        query_points = positions.array.transpose(0, 2, 1).reshape(-1, 2)
        _, nearest_indices = tree.query(query_points)
        nearest_arc_lengths = arc_lengths[nearest_indices].reshape(T, M)

        return nearest_arc_lengths

    def _refine(
        self,
        arc_lengths: Float[Array, "T M"],
        positions: NumPyPositions,
    ) -> Float[Array, "T M"]:
        # Refines the arc lengths using Newton's method.
        s_0 = arc_lengths
        x, y = positions.x(), positions.y()

        def f(s):
            return (x - self.spline_x(s)) * self.spline_x(s, 1) + (
                y - self.spline_y(s)
            ) * self.spline_y(s, 1)

        def f_prime(s):
            c_x, c_y = self.spline_x(s), self.spline_y(s)
            dx, dy = self.spline_x(s, 1), self.spline_y(s, 1)
            ddx, ddy = self.spline_x(s, 2), self.spline_y(s, 2)
            return -(dx**2 + dy**2) + (x - c_x) * ddx + (y - c_y) * ddy

        # NOTE: Newton may not fully converge for points far from trajectory.
        # This is fine, since we don't care about high precision for such points.
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", message="some failed to converge")
            result = newton(
                f, s_0, fprime=f_prime, maxiter=self.refining_iterations, disp=False
            )

        return np.clip(result, 0.0, self.path_length)

    @cached_property
    def _guess_samples(self) -> tuple[Float[Array, " _"], cKDTree]:
        # Computes samples along the spline and builds a KD-tree for fast lookup.
        s = np.linspace(0, self.path_length, self.kd_tree_samples)

        return s, cKDTree(np.column_stack([self.spline_x(s), self.spline_y(s)]))


def compute_path_parameters(
    *, x: Float[Array, " L"], y: Float[Array, " L"]
) -> Float[Array, " L"]:
    segment_lengths = np.sqrt(np.diff(x) ** 2 + np.diff(y) ** 2)
    return np.concatenate([[0.0], np.cumsum(segment_lengths)])



from typing import Self
from dataclasses import dataclass

from faran.types import jaxtyped, JaxObstacleSimulator
from faran.obstacles.accelerated import JaxObstacle2dPosesForTimeStep

from jaxtyping import Array as JaxArray, Float, Scalar

import jax
import jax.numpy as jnp


@dataclass(kw_only=True)
class JaxDynamicObstacleSimulator(JaxObstacleSimulator[JaxObstacle2dPosesForTimeStep]):
    """Simulates obstacles moving with constant velocity over the prediction horizon."""

    last: JaxObstacle2dPosesForTimeStep
    velocities: Float[JaxArray, "K 2"]

    time_step: Scalar | None = None

    @staticmethod
    def create(
        *, positions: Float[JaxArray, "K 2"], velocities: Float[JaxArray, "K 2"]
    ) -> "JaxDynamicObstacleSimulator":
        headings = headings_from(velocities)

        return JaxDynamicObstacleSimulator(
            last=JaxObstacle2dPosesForTimeStep.create(
                x=positions[:, 0], y=positions[:, 1], heading=headings
            ),
            velocities=velocities,
        )

    def with_time_step_size(self, time_step_size: float) -> Self:
        return self.__class__(
            last=self.last,
            velocities=self.velocities,
            time_step=jnp.asarray(time_step_size),
        )

    def step(self) -> JaxObstacle2dPosesForTimeStep:
        assert self.time_step is not None, (
            "Time step must be set to advance obstacle states."
        )

        x, y = step_obstacles(
            x=self.last.x_array,
            y=self.last.y_array,
            velocities=self.velocities,
            time_step=self.time_step,
        )

        self.last = JaxObstacle2dPosesForTimeStep.create(
            x=x, y=y, heading=self.last.heading_array
        )

        return self.last

    @property
    def obstacle_count(self) -> int:
        return self.last.count


@jax.jit
@jaxtyped
def step_obstacles(
    *,
    x: Float[JaxArray, " K"],
    y: Float[JaxArray, " K"],
    velocities: Float[JaxArray, "K 2"],
    time_step: Scalar,
) -> tuple[Float[JaxArray, " K"], Float[JaxArray, " K"]]:
    new_x = x + velocities[:, 0] * time_step
    new_y = y + velocities[:, 1] * time_step
    return new_x, new_y


@jax.jit
@jaxtyped
def headings_from(velocities: Float[JaxArray, "K 2"]) -> Float[JaxArray, " K"]:
    speed = jnp.linalg.norm(velocities, axis=1)
    return jnp.where(speed > 1e-6, jnp.arctan2(velocities[:, 1], velocities[:, 0]), 0.0)



from typing import overload
from dataclasses import dataclass

from faran.types import (
    DistanceExtractor,
    Array,
    NumPyHeadings,
    NumPyPositions,
    NumPyPositionExtractor,
    NumPyHeadingExtractor,
    NumPySampledObstaclePositions,
    NumPySampledObstacleHeadings,
    NumPySampledObstaclePositionExtractor,
    NumPySampledObstacleHeadingExtractor,
)
from faran.costs.collision import NumPyDistance
from faran.costs.distance.basic import replace_missing
from faran.costs.distance.circles.common import Circles

from jaxtyping import Float

import numpy as np


type OriginsArray = Float[Array, "N 2"]
type RadiiArray = Float[Array, " N"]


@dataclass(frozen=True)
class NumPyCircleDistanceExtractor[StateT, SampledObstacleStatesT](
    DistanceExtractor[StateT, SampledObstacleStatesT, NumPyDistance]
):
    """
    Computes the distances between parts of the ego robot and obstacles. Both the ego
    and the obstacles are represented as collections of circles.
    """

    ego: Circles
    obstacle: Circles
    positions_from: NumPyPositionExtractor[StateT]
    headings_from: NumPyHeadingExtractor[StateT]
    obstacle_positions_from: NumPySampledObstaclePositionExtractor[
        SampledObstacleStatesT
    ]
    obstacle_headings_from: NumPySampledObstacleHeadingExtractor[SampledObstacleStatesT]

    @staticmethod
    def create[S, SOS](
        *,
        ego: Circles,
        obstacle: Circles,
        position_extractor: NumPyPositionExtractor[S],
        heading_extractor: NumPyHeadingExtractor[S],
        obstacle_position_extractor: NumPySampledObstaclePositionExtractor[SOS],
        obstacle_heading_extractor: NumPySampledObstacleHeadingExtractor[SOS],
    ) -> "NumPyCircleDistanceExtractor[S, SOS]":
        return NumPyCircleDistanceExtractor(
            ego=ego,
            obstacle=obstacle,
            positions_from=position_extractor,
            headings_from=heading_extractor,
            obstacle_positions_from=obstacle_position_extractor,
            obstacle_headings_from=obstacle_heading_extractor,
        )

    def __call__(
        self, *, states: StateT, obstacle_states: SampledObstacleStatesT
    ) -> NumPyDistance:
        obstacle_positions, obstacle_headings = replace_missing(
            positions=self.obstacle_positions_from(obstacle_states),
            headings=self.obstacle_headings_from(obstacle_states),
        )

        return NumPyDistance(
            compute_circle_distances(
                ego_positions=self.positions_from(states),
                ego_headings=self.headings_from(states),
                ego=self.ego,
                obstacle_positions=obstacle_positions,
                obstacle_headings=obstacle_headings,
                obstacle=self.obstacle,
            )
        )


def compute_circle_distances(
    *,
    ego_positions: NumPyPositions,
    ego_headings: NumPyHeadings,
    ego: Circles,
    obstacle_positions: NumPySampledObstaclePositions,
    obstacle_headings: NumPySampledObstacleHeadings,
    obstacle: Circles,
) -> Float[Array, "T V M N"]:
    ego_global_x, ego_global_y = to_global_positions(
        x=ego_positions.x(),
        y=ego_positions.y(),
        heading=ego_headings.heading(),
        local_origins=ego.origins,
    )

    obstacle_global_x, obstacle_global_y = to_global_positions(
        x=obstacle_positions.x(),
        y=obstacle_positions.y(),
        heading=obstacle_headings.heading(),
        local_origins=obstacle.origins,
    )

    pairwise_distances = pairwise_min_distances(
        ego_x=ego_global_x,
        ego_y=ego_global_y,
        ego_radii=ego.radii,
        obstacle_x=obstacle_global_x,
        obstacle_y=obstacle_global_y,
        obstacle_radii=obstacle.radii,
    )

    return min_distance_per_ego_part(pairwise_distances)


@overload
def to_global_positions(
    *,
    x: Float[Array, "T M"],
    y: Float[Array, "T M"],
    heading: Float[Array, "T M"],
    local_origins: OriginsArray,
) -> tuple[Float[Array, "V T M"], Float[Array, "V T M"]]: ...


@overload
def to_global_positions(  # pyright: ignore[reportOverlappingOverload]
    *,
    x: Float[Array, "T K N"],
    y: Float[Array, "T K N"],
    heading: Float[Array, "T K N"],
    local_origins: OriginsArray,
) -> tuple[Float[Array, "C T K N"], Float[Array, "C T K N"]]: ...


def to_global_positions(
    *,
    x: Float[Array, "T M"] | Float[Array, "T K N"],
    y: Float[Array, "T M"] | Float[Array, "T K N"],
    heading: Float[Array, "T M"] | Float[Array, "T K N"],
    local_origins: OriginsArray | OriginsArray,
) -> (
    tuple[Float[Array, "V T M"], Float[Array, "V T M"]]
    | tuple[Float[Array, "C T K N"], Float[Array, "C T K N"]]
):
    local_xy = local_origins.reshape((-1, 2) + (1,) * x.ndim)
    local_x, local_y = local_xy[:, 0], local_xy[:, 1]

    cos_h, sin_h = np.cos(heading), np.sin(heading)

    return (
        x + local_x * cos_h - local_y * sin_h,
        y + local_x * sin_h + local_y * cos_h,
    )


def pairwise_min_distances(
    ego_x: Float[Array, "V T M"],
    ego_y: Float[Array, "V T M"],
    ego_radii: RadiiArray,
    obstacle_x: Float[Array, "C T K N"],
    obstacle_y: Float[Array, "C T K N"],
    obstacle_radii: RadiiArray,
) -> Float[Array, "V C T M K N"]:
    dx = (
        ego_x[:, np.newaxis, :, :, np.newaxis, np.newaxis]
        - obstacle_x[np.newaxis, :, :, np.newaxis, :, :]
    )
    dy = (
        ego_y[:, np.newaxis, :, :, np.newaxis, np.newaxis]
        - obstacle_y[np.newaxis, :, :, np.newaxis, :, :]
    )

    center_distances = np.sqrt(dx**2 + dy**2)
    radii_sum = (
        ego_radii[:, np.newaxis, np.newaxis, np.newaxis, np.newaxis, np.newaxis]
        + obstacle_radii[np.newaxis, :, np.newaxis, np.newaxis, np.newaxis, np.newaxis]
    )

    return center_distances - radii_sum


def min_distance_per_ego_part(
    pairwise_distances: Float[Array, "V C T M K N"],
) -> Float[Array, "T V M N"]:
    V, C, T, M, K, N = pairwise_distances.shape

    if C == 0 or K == 0:
        return np.full((T, V, M, N), np.inf)

    min_over_obstacles = np.min(pairwise_distances, axis=(1, 4))
    return np.transpose(min_over_obstacles, (1, 0, 2, 3))



from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    jaxtyped,
    Array,
    Trajectory,
    NumPyPathParameters,
    NumPyReferencePoints,
    NumPyPositions,
    NumPyLateralPositions,
    NumPyLongitudinalPositions,
    NumPyNormals,
)

from jaxtyping import Float

import numpy as np


type Vector = Float[Array, "2"]


@jaxtyped
@dataclass(kw_only=True, frozen=True)
class NumPyLineTrajectory(
    Trajectory[
        NumPyPathParameters,
        NumPyReferencePoints,
        NumPyPositions,
        NumPyLateralPositions,
        NumPyLongitudinalPositions,
    ]
):
    """NumPy straight-line reference trajectory from start to end."""

    start: Vector
    direction: Vector

    heading: float

    _end: tuple[float, float]
    _path_length: float

    @staticmethod
    def create(
        *,
        start: tuple[float, float],
        end: tuple[float, float],
        path_length: float | None = None,
    ) -> "NumPyLineTrajectory":
        """Generates a straight line trajectory from start to end.

        Args:
            start: The starting point of the trajectory.
            end: The ending point of the trajectory.
            path_length: Total length of the trajectory. If omitted, it will be set to the natural length
                of the line segment between start and end.
        """
        return NumPyLineTrajectory(
            start=(start_array := np.array(start)),
            direction=(direction := np.array(end) - start_array),
            heading=np.arctan2(direction[1], direction[0]),
            _end=end,
            _path_length=length_of(direction) if path_length is None else path_length,
        )

    def query(self, parameters: NumPyPathParameters) -> NumPyReferencePoints:
        normalized = parameters.array / self.path_length

        x, y = (
            self.start[:, np.newaxis, np.newaxis]
            + self.direction[:, np.newaxis, np.newaxis] * normalized
        )
        heading = np.full_like(x, self.heading)

        return NumPyReferencePoints.create(x=x, y=y, heading=heading)

    def lateral(self, positions: NumPyPositions) -> NumPyLateralPositions:
        relative = positions.array - self.start[:, np.newaxis]
        lateral = np.einsum("tpm,p->tm", relative, self.perpendicular)

        return NumPyLateralPositions.create(lateral)

    def longitudinal(self, positions: NumPyPositions) -> NumPyLongitudinalPositions:
        relative = positions.array - self.start[:, np.newaxis]
        projection = np.einsum("tpm,p->tm", relative, self.tangent)
        longitudinal = projection * self.path_length / self.line_length

        return NumPyLongitudinalPositions.create(longitudinal)

    def normal(self, parameters: NumPyPathParameters) -> NumPyNormals:
        T, M = parameters.horizon, parameters.rollout_count

        x = np.full((T, M), self.perpendicular[0])
        y = np.full((T, M), self.perpendicular[1])

        return NumPyNormals.create(x=x, y=y)

    @property
    def end(self) -> tuple[float, float]:
        return self._end

    @property
    def path_length(self) -> float:
        return self._path_length

    @property
    def natural_length(self) -> float:
        return self.line_length

    @cached_property
    def perpendicular(self) -> Vector:
        tangent = self.tangent
        perpendicular = np.array([tangent[1], -tangent[0]])

        return perpendicular

    @cached_property
    def tangent(self) -> Vector:
        tangent = self.direction / self.line_length

        return tangent

    @cached_property
    def line_length(self) -> float:
        return length_of(self.direction)


def length_of(line: Vector) -> float:
    return float(np.linalg.norm(line))



from .common import Circles as Circles
from .basic import NumPyCircleDistanceExtractor as NumPyCircleDistanceExtractor
from .accelerated import JaxCircleDistanceExtractor as JaxCircleDistanceExtractor



from .basic import (
    NumPyBicycleState as NumPyBicycleState,
    NumPyBicycleStateSequence as NumPyBicycleStateSequence,
    NumPyBicycleStateBatch as NumPyBicycleStateBatch,
    NumPyBicyclePositions as NumPyBicyclePositions,
    NumPyBicycleControlInputSequence as NumPyBicycleControlInputSequence,
    NumPyBicycleControlInputBatch as NumPyBicycleControlInputBatch,
    NumPyBicycleObstacleStates as NumPyBicycleObstacleStates,
    NumPyBicycleObstacleInputs as NumPyBicycleObstacleInputs,
    NumPyBicycleObstacleStateSequences as NumPyBicycleObstacleStateSequences,
    NumPyBicycleModel as NumPyBicycleModel,
    NumPyBicycleObstacleModel as NumPyBicycleObstacleModel,
    NumPyFiniteDifferenceBicycleStateEstimator as NumPyFiniteDifferenceBicycleStateEstimator,
    NumPyKfBicycleStateEstimator as NumPyKfBicycleStateEstimator,
)
from .accelerated import (
    JaxBicycleState as JaxBicycleState,
    JaxBicycleStateSequence as JaxBicycleStateSequence,
    JaxBicycleStateBatch as JaxBicycleStateBatch,
    JaxBicyclePositions as JaxBicyclePositions,
    JaxBicycleControlInputSequence as JaxBicycleControlInputSequence,
    JaxBicycleControlInputBatch as JaxBicycleControlInputBatch,
    JaxBicycleObstacleStates as JaxBicycleObstacleStates,
    JaxBicycleObstacleInputs as JaxBicycleObstacleInputs,
    JaxBicycleObstacleStateSequences as JaxBicycleObstacleStateSequences,
    JaxBicycleModel as JaxBicycleModel,
    JaxBicycleObstacleModel as JaxBicycleObstacleModel,
    JaxFiniteDifferenceBicycleStateEstimator as JaxFiniteDifferenceBicycleStateEstimator,
    JaxKfBicycleStateEstimator as JaxKfBicycleStateEstimator,
)



from typing import Self, cast, Sequence, Final
from dataclasses import dataclass
from functools import cached_property

from faran.types import (
    jaxtyped,
    Array,
    DataType,
    NumPyState,
    NumPyStateSequence,
    NumPyStateBatch,
    NumPyControlInputSequence,
    NumPyControlInputBatch,
    NumPyBicycleObstacleStatesHistory,
    BicycleState,
    BicycleStateSequence,
    BicycleStateBatch,
    BicycleControlInputSequence,
    BicycleControlInputBatch,
    BicyclePositions,
    BicycleD_x,
    BICYCLE_D_X,
    BicycleD_u,
    BICYCLE_D_U,
    BicycleD_o,
    DynamicalModel,
    ObstacleModel,
    ObstacleStateEstimator,
    EstimatedObstacleStates,
    NumPyGaussianBelief,
    NumPyNoiseModelProvider,
    NumPyNoiseCovarianceDescription,
)
from faran.filters import (
    NumPyExtendedKalmanFilter,
    NumPyUnscentedKalmanFilter,
    numpy_kalman_filter,
)
from faran.obstacles import NumPyObstacle2dPoses
from faran.models.common import SMALL_UNCERTAINTY, LARGE_UNCERTAINTY
from faran.models.basic import invalid_obstacle_filter_from

from numtypes import D, array

from jaxtyping import Float

import numpy as np


NO_LIMITS: Final = (float("-inf"), float("inf"))

BICYCLE_ESTIMATION_D_X: Final = 6
BICYCLE_OBSERVATION_D_O: Final = 3

type BicycleEstimationD_x = D[6]
type BicycleObservationD_o = D[3]

type StateArray = Float[Array, " BicycleD_x"]
type ControlInputSequenceArray = Float[Array, "T BicycleD_u"]
type StateBatchArray = Float[Array, "T BicycleD_x M"]
type ControlInputBatchArray = Float[Array, "T BicycleD_u M"]

type StatesAtTimeStep = Float[Array, "BicycleD_x M"]
type ControlInputsAtTimeStep = Float[Array, "BicycleD_u M"]

type EstimationStateCovarianceArray = Float[
    Array, "BicycleEstimationD_x BicycleEstimationD_x"
]
type ProcessNoiseCovarianceArray = Float[
    Array, "BicycleEstimationD_x BicycleEstimationD_x"
]
type ObservationNoiseCovarianceArray = Float[
    Array, "BicycleObservationD_o BicycleObservationD_o"
]
type ObservationMatrix = Float[Array, "BicycleObservationD_o BicycleEstimationD_x"]

type BicycleGaussianBelief = NumPyGaussianBelief
type NumPyBicycleObstacleCovariances = Float[
    Array, "BicycleEstimationD_x BicycleEstimationD_x K"
]
type KalmanFilter = NumPyExtendedKalmanFilter | NumPyUnscentedKalmanFilter


@jaxtyped
@dataclass(frozen=True)
class NumPyBicycleState(BicycleState, NumPyState):
    """Kinematic bicycle state: [x, y, heading, speed]."""

    _array: StateArray

    @staticmethod
    def create(
        *, x: float, y: float, heading: float, speed: float
    ) -> "NumPyBicycleState":
        return NumPyBicycleState(array([x, y, heading, speed], shape=(BICYCLE_D_X,)))

    def __array__(self, dtype: DataType | None = None) -> StateArray:
        return self.array

    @property
    def dimension(self) -> BicycleD_x:
        return cast(BicycleD_x, self.array.shape[0])

    @property
    def x(self) -> float:
        return self.array[0]

    @property
    def y(self) -> float:
        return self.array[1]

    @property
    def heading(self) -> float:
        return self.array[2]

    @property
    def speed(self) -> float:
        return self.array[3]

    @property
    def array(self) -> StateArray:
        return self._array


@dataclass(kw_only=True, frozen=True)
class NumPyBicycleStateSequence(BicycleStateSequence, NumPyStateSequence):
    batch: "NumPyBicycleStateBatch"
    rollout: int

    @staticmethod
    def of_states(states: Sequence[NumPyBicycleState]) -> "NumPyBicycleStateSequence":
        assert len(states) > 0, "States sequence must not be empty."

        array = np.stack([state.array for state in states], axis=0)[:, :, np.newaxis]

        return NumPyBicycleStateSequence(
            batch=NumPyBicycleStateBatch.wrap(array), rollout=0
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T BicycleD_x"]:
        return self.array

    def step(self, index: int) -> NumPyBicycleState:
        return NumPyBicycleState(self.array[index])

    def batched(self) -> "NumPyBicycleStateBatch":
        return NumPyBicycleStateBatch.wrap(self.array[..., np.newaxis])

    def x(self) -> Float[Array, " T"]:
        return self.array[:, 0]

    def y(self) -> Float[Array, " T"]:
        return self.array[:, 1]

    def heading(self) -> Float[Array, " T"]:
        return self.array[:, 2]

    def speed(self) -> Float[Array, " T"]:
        return self.array[:, 3]

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_x:
        return cast(BicycleD_x, self.array.shape[1])

    @property
    def array(self) -> Float[Array, "T BicycleD_x"]:
        return self.batch.array[:, :, self.rollout]


@jaxtyped
@dataclass(frozen=True)
class NumPyBicycleStateBatch(BicycleStateBatch, NumPyStateBatch):
    _array: StateBatchArray

    @staticmethod
    def wrap(
        array: StateBatchArray,
    ) -> "NumPyBicycleStateBatch":
        return NumPyBicycleStateBatch(array)

    @staticmethod
    def of_states(states: Sequence[NumPyBicycleState]) -> "NumPyBicycleStateBatch":
        assert len(states) > 0, "States sequence must not be empty."

        array = np.stack([state.array for state in states], axis=0)[:, :, np.newaxis]

        return NumPyBicycleStateBatch(array)

    def __array__(self, dtype: DataType | None = None) -> StateBatchArray:
        return self.array

    def heading(self) -> Float[Array, "T M"]:
        return self.array[:, 2, :]

    def speed(self) -> Float[Array, "T M"]:
        return self.array[:, 3, :]

    def rollout(self, index: int) -> NumPyBicycleStateSequence:
        return NumPyBicycleStateSequence(batch=self, rollout=index)

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_x:
        return cast(BicycleD_x, self.array.shape[1])

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def positions(self) -> "NumPyBicyclePositions":
        return NumPyBicyclePositions(batch=self)

    @property
    def array(self) -> StateBatchArray:
        return self._array


@dataclass(frozen=True)
class NumPyBicyclePositions(BicyclePositions):
    batch: NumPyBicycleStateBatch

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T 2 M"]:
        return self.batch.array[:, :2, :]

    def x(self) -> Float[Array, "T M"]:
        return self.batch.array[:, 0, :]

    def y(self) -> Float[Array, "T M"]:
        return self.batch.array[:, 1, :]

    @property
    def array(self) -> Float[Array, "T 2 M"]:
        return self.batch.array[:, :2, :]

    @property
    def horizon(self) -> int:
        return self.batch.horizon

    @property
    def dimension(self) -> D[2]:
        return 2

    @property
    def rollout_count(self) -> int:
        return self.batch.rollout_count


@jaxtyped
@dataclass(frozen=True)
class NumPyBicycleControlInputSequence(
    BicycleControlInputSequence, NumPyControlInputSequence
):
    """Control inputs: [acceleration, steering_angle]."""

    _array: ControlInputSequenceArray

    @staticmethod
    def zeroes(horizon: int) -> "NumPyBicycleControlInputSequence":
        """Creates a zeroed control input sequence for the given horizon."""
        return NumPyBicycleControlInputSequence(np.zeros((horizon, BICYCLE_D_U)))

    def __array__(self, dtype: DataType | None = None) -> ControlInputSequenceArray:
        return self.array

    def similar(self, *, array: Float[Array, "L BicycleD_u"]) -> Self:
        return self.__class__(array)

    def accelerations(self) -> Float[Array, " T"]:
        return self.array[:, 0]

    def steering_angles(self) -> Float[Array, " T"]:
        return self.array[:, 1]

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_u:
        return cast(BicycleD_u, self.array.shape[1])

    @property
    def array(self) -> ControlInputSequenceArray:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class NumPyBicycleControlInputBatch(BicycleControlInputBatch, NumPyControlInputBatch):
    _array: ControlInputBatchArray

    @staticmethod
    def zero(
        *, horizon: int, rollout_count: int = 1
    ) -> "NumPyBicycleControlInputBatch":
        return NumPyBicycleControlInputBatch(
            np.zeros((horizon, BICYCLE_D_U, rollout_count))
        )

    @staticmethod
    def create(
        *, array: Float[Array, "T BicycleD_u M"]
    ) -> "NumPyBicycleControlInputBatch":
        return NumPyBicycleControlInputBatch(array)

    @staticmethod
    def of(
        sequence: NumPyBicycleControlInputSequence,
    ) -> "NumPyBicycleControlInputBatch":
        return NumPyBicycleControlInputBatch(sequence.array[..., np.newaxis])

    def __array__(self, dtype: DataType | None = None) -> ControlInputBatchArray:
        return self.array

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleD_u:
        return cast(BicycleD_u, self.array.shape[1])

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def array(self) -> ControlInputBatchArray:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class NumPyBicycleObstacleStates:
    _array: Float[Array, "BicycleD_o K"]

    @staticmethod
    def wrap(
        array: Float[Array, "BicycleD_o K"],
    ) -> "NumPyBicycleObstacleStates":
        return NumPyBicycleObstacleStates(array)

    @staticmethod
    def create(
        *,
        x: Float[Array, " K"],
        y: Float[Array, " K"],
        heading: Float[Array, " K"],
        speed: Float[Array, " K"],
    ) -> "NumPyBicycleObstacleStates":
        return NumPyBicycleObstacleStates(np.stack([x, y, heading, speed], axis=0))

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "BicycleD_o K"]:
        return self.array

    def x(self) -> Float[Array, " K"]:
        return self.array[0, :]

    def y(self) -> Float[Array, " K"]:
        return self.array[1, :]

    def heading(self) -> Float[Array, " K"]:
        return self.array[2, :]

    def speed(self) -> Float[Array, " K"]:
        return self.array[3, :]

    @property
    def dimension(self) -> BicycleD_o:
        return cast(BicycleD_o, self.array.shape[0])

    @property
    def count(self) -> int:
        return self.array.shape[1]

    @property
    def array(self) -> Float[Array, "BicycleD_o K"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class NumPyBicycleObstacleStateSequences:
    _array: Float[Array, "T BicycleEstimationD_x K"]
    _covariance: Float[Array, "T BicycleEstimationD_x BicycleEstimationD_x K"]

    @staticmethod
    def create(
        predictions: Sequence[BicycleGaussianBelief],
    ) -> "NumPyBicycleObstacleStateSequences":
        assert len(predictions) > 0, "Predictions sequence must not be empty."

        T_ = len(predictions)
        K_ = predictions[0].mean.shape[1]

        array = np.stack([belief.mean for belief in predictions], axis=0)
        covariance = np.stack([belief.covariance for belief in predictions], axis=0)

        assert array.shape == (T_, BICYCLE_ESTIMATION_D_X, K_)
        assert covariance.shape == (
            T_,
            BICYCLE_ESTIMATION_D_X,
            BICYCLE_ESTIMATION_D_X,
            K_,
        )

        return NumPyBicycleObstacleStateSequences(array, covariance)

    def __array__(
        self, dtype: DataType | None = None
    ) -> Float[Array, "T BicycleEstimationD_x K"]:
        return self.array

    def x(self) -> Float[Array, "T K"]:
        return self.array[:, 0, :]

    def y(self) -> Float[Array, "T K"]:
        return self.array[:, 1, :]

    def heading(self) -> Float[Array, "T K"]:
        return self.array[:, 2, :]

    def covariance(
        self,
    ) -> Float[Array, "T BicycleEstimationD_x BicycleEstimationD_x K"]:
        return self._covariance

    def pose_covariance(self) -> Float[Array, "T PoseD_o PoseD_o K"]:
        return self._covariance[:, :3, :3, :]

    def pose(self) -> NumPyObstacle2dPoses:
        return NumPyObstacle2dPoses.create(
            x=self.x(),
            y=self.y(),
            heading=self.heading(),
            covariance=self.pose_covariance(),
        )

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def dimension(self) -> BicycleEstimationD_x:
        return cast(BicycleEstimationD_x, self.array.shape[1])

    @property
    def count(self) -> int:
        return self.array.shape[2]

    @property
    def array(self) -> Float[Array, "T BicycleEstimationD_x K"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class NumPyBicycleObstacleInputs:
    _accelerations: Float[Array, " K"]
    _steering_angles: Float[Array, " K"]

    @staticmethod
    def wrap(inputs: Float[Array, "BicycleD_u K"]) -> "NumPyBicycleObstacleInputs":
        accelerations, steering_angles = inputs
        return NumPyBicycleObstacleInputs(
            _accelerations=accelerations, _steering_angles=steering_angles
        )

    @staticmethod
    def create(
        *, accelerations: Float[Array, " K"], steering_angles: Float[Array, " K"]
    ) -> "NumPyBicycleObstacleInputs":
        return NumPyBicycleObstacleInputs(
            _accelerations=accelerations,
            _steering_angles=steering_angles,
        )

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "BicycleD_u K"]:
        return self.array

    def accelerations(self) -> Float[Array, " K"]:
        return self._accelerations

    def steering_angles(self) -> Float[Array, " K"]:
        return self._steering_angles

    def zeroed(
        self, *, acceleration: bool = False, steering_angle: bool = False
    ) -> "NumPyBicycleObstacleInputs":
        """Returns a version of the inputs with the specified components zeroed out."""
        return NumPyBicycleObstacleInputs(
            _accelerations=np.zeros_like(self._accelerations)
            if acceleration
            else self._accelerations,
            _steering_angles=np.zeros_like(self._steering_angles)
            if steering_angle
            else self._steering_angles,
        )

    @property
    def dimension(self) -> BicycleD_u:
        return BICYCLE_D_U

    @property
    def count(self) -> int:
        return self._steering_angles.shape[0]

    @property
    def array(self) -> Float[Array, "BicycleD_u K"]:
        return self._array

    @cached_property
    def _array(self) -> Float[Array, "BicycleD_u K"]:
        return np.stack([self._accelerations, self._steering_angles], axis=0)


@dataclass(kw_only=True, frozen=True)
class NumPyBicycleModel(
    DynamicalModel[
        NumPyBicycleState,
        NumPyBicycleStateSequence,
        NumPyBicycleStateBatch,
        NumPyBicycleControlInputSequence,
        NumPyBicycleControlInputBatch,
    ],
):
    """Kinematic bicycle with rear-axle reference, Euler-integrated with configurable limits."""

    _time_step_size: float
    wheelbase: float
    rear_axle_distance: float
    speed_limits: tuple[float, float]
    steering_limits: tuple[float, float]
    acceleration_limits: tuple[float, float]

    @staticmethod
    def create(
        *,
        time_step_size: float,
        wheelbase: float = 1.0,
        rear_axle_distance: float = 0.0,
        speed_limits: tuple[float, float] | None = None,
        steering_limits: tuple[float, float] | None = None,
        acceleration_limits: tuple[float, float] | None = None,
    ) -> "NumPyBicycleModel":

        return NumPyBicycleModel(
            _time_step_size=time_step_size,
            wheelbase=wheelbase,
            rear_axle_distance=rear_axle_distance,
            speed_limits=speed_limits if speed_limits is not None else NO_LIMITS,
            steering_limits=steering_limits
            if steering_limits is not None
            else NO_LIMITS,
            acceleration_limits=acceleration_limits
            if acceleration_limits is not None
            else NO_LIMITS,
        )

    def simulate(
        self, inputs: NumPyBicycleControlInputBatch, initial_state: NumPyBicycleState
    ) -> NumPyBicycleStateBatch:
        rollout_count = inputs.rollout_count

        initial = np.stack(
            [
                np.full(rollout_count, initial_state.x),
                np.full(rollout_count, initial_state.y),
                np.full(rollout_count, initial_state.heading),
                np.full(rollout_count, initial_state.speed),
            ]
        )

        return NumPyBicycleStateBatch(
            simulate(
                inputs.array,
                initial,
                time_step_size=self.time_step_size,
                wheelbase=self.wheelbase,
                rear_axle_distance=self.rear_axle_distance,
                speed_limits=self.speed_limits,
                steering_limits=self.steering_limits,
                acceleration_limits=self.acceleration_limits,
            )
        )

    def step(
        self, inputs: NumPyBicycleControlInputSequence, state: NumPyBicycleState
    ) -> NumPyBicycleState:
        state_as_rollouts = state.array.reshape(-1, 1)
        first_input = inputs.array[0].reshape(-1, 1)

        return NumPyBicycleState(
            step(
                state_as_rollouts,
                first_input,
                time_step_size=self.time_step_size,
                wheelbase=self.wheelbase,
                rear_axle_distance=self.rear_axle_distance,
                speed_limits=self.speed_limits,
                steering_limits=self.steering_limits,
                acceleration_limits=self.acceleration_limits,
            )[:, 0]
        )

    def forward(
        self, inputs: NumPyBicycleControlInputSequence, state: NumPyBicycleState
    ) -> NumPyBicycleStateSequence:
        return self.simulate(NumPyBicycleControlInputBatch.of(inputs), state).rollout(0)

    @property
    def time_step_size(self) -> float:
        return self._time_step_size


@dataclass(frozen=True)
class NumPyBicycleStateEstimationModel:
    """Kinematic bicycle model used for state estimation."""

    time_step_size: float
    wheelbase: float
    rear_axle_distance: float
    initial_state_covariance: EstimationStateCovarianceArray

    @staticmethod
    def create(
        *,
        time_step_size: float,
        wheelbase: float,
        rear_axle_distance: float = 0.0,
        initial_state_covariance: EstimationStateCovarianceArray | None = None,
    ) -> "NumPyBicycleStateEstimationModel":
        if initial_state_covariance is None:
            initial_state_covariance = (
                NumPyBicycleStateEstimationModel.default_initial_state_covariance()
            )

        return NumPyBicycleStateEstimationModel(
            time_step_size=time_step_size,
            wheelbase=wheelbase,
            rear_axle_distance=rear_axle_distance,
            initial_state_covariance=initial_state_covariance,
        )

    @staticmethod
    def default_initial_state_covariance() -> EstimationStateCovarianceArray:
        # NOTE: Sure of pose, unsure of everything else.
        # Steering angle covariance is kept somewhat smaller (0.1), since it
        # otherwise messes up UKF.
        return cast(
            EstimationStateCovarianceArray,
            np.diag(
                [
                    SMALL_UNCERTAINTY,
                    SMALL_UNCERTAINTY,
                    SMALL_UNCERTAINTY,
                    LARGE_UNCERTAINTY,
                    LARGE_UNCERTAINTY,
                    0.1,
                ]
            ),
        )

    def __call__(self, state: Float[Array, "D_x K"]) -> Float[Array, "D_x K"]:
        dt = self.time_step_size
        l_r = self.rear_axle_distance
        x, y, theta, v, a, delta = state

        beta = np.arctan(l_r / self.wheelbase * np.tan(delta))

        return np.array(
            [
                x + v * np.cos(theta + beta) * dt,
                y + v * np.sin(theta + beta) * dt,
                theta + v * np.cos(beta) * np.tan(delta) / self.wheelbase * dt,
                v + a * dt,
                a,
                delta,
            ]
        )

    def jacobian(self, state: Float[Array, "D_x K"]) -> Float[Array, "D_x D_x K"]:
        D_x, K = state.shape
        dt = self.time_step_size
        l_r = self.rear_axle_distance
        L = self.wheelbase
        x, y, theta, v, a, delta = state

        tan_delta = np.tan(delta)
        sec_delta_squared = 1 / np.cos(delta) ** 2

        beta = np.arctan(l_r / L * tan_delta)
        cos_beta = np.cos(beta)
        sin_beta = np.sin(beta)
        cos_theta_beta = np.cos(theta + beta)
        sin_theta_beta = np.sin(theta + beta)

        dbeta_ddelta = (l_r / L * sec_delta_squared) / (1 + (l_r / L * tan_delta) ** 2)

        jacobian = np.zeros((D_x, D_x, K))

        # Partial derivatives for x
        jacobian[0, 0, :] = 1  # ∂x_next/∂x
        jacobian[0, 2, :] = -v * sin_theta_beta * dt  # ∂x_next/∂theta
        jacobian[0, 3, :] = cos_theta_beta * dt  # ∂x_next/∂v
        jacobian[0, 5, :] = -v * sin_theta_beta * dbeta_ddelta * dt  # ∂x_next/∂delta

        # Partial derivatives for y
        jacobian[1, 1, :] = 1  # ∂y_next/∂y
        jacobian[1, 2, :] = v * cos_theta_beta * dt  # ∂y_next/∂theta
        jacobian[1, 3, :] = sin_theta_beta * dt  # ∂y_next/∂v
        jacobian[1, 5, :] = v * cos_theta_beta * dbeta_ddelta * dt  # ∂y_next/∂delta

        # Partial derivatives for theta
        jacobian[2, 2, :] = 1  # ∂theta_next/∂theta
        jacobian[2, 3, :] = cos_beta * tan_delta / L * dt  # ∂theta_next/∂v
        jacobian[2, 5, :] = (
            v
            / L
            * (cos_beta * sec_delta_squared - sin_beta * dbeta_ddelta * tan_delta)
            * dt
        )  # ∂theta_next/∂delta

        # Partial derivatives for v
        jacobian[3, 3, :] = 1  # ∂v_next/∂v
        jacobian[3, 4, :] = dt  # ∂v_next/∂a

        # Partial derivatives for a and delta are identity since they are assumed constant
        jacobian[4, 4, :] = 1  # ∂a_next/∂a
        jacobian[5, 5, :] = 1  # ∂delta_next/∂delta

        return jacobian

    def observations_from(
        self, history: NumPyBicycleObstacleStatesHistory
    ) -> Float[Array, "T BicycleObservationD_o K"]:
        return np.stack([history.x(), history.y(), history.heading()], axis=1)

    def states_from(self, belief: BicycleGaussianBelief) -> NumPyBicycleObstacleStates:
        return NumPyBicycleObstacleStates.wrap(belief.mean[:4, :])

    def inputs_from(self, belief: BicycleGaussianBelief) -> NumPyBicycleObstacleInputs:
        return NumPyBicycleObstacleInputs.wrap(belief.mean[4:, :])

    def initial_belief_from(
        self,
        *,
        states: NumPyBicycleObstacleStates,
        inputs: NumPyBicycleObstacleInputs,
        covariances: NumPyBicycleObstacleCovariances | None = None,
    ) -> BicycleGaussianBelief:
        augmented = np.concatenate([states.array, inputs.array], axis=0)

        if covariances is None:
            # NOTE: No covariance means we are "certain" about the states.
            covariances = np.broadcast_to(  # type: ignore
                np.eye(BICYCLE_ESTIMATION_D_X)[:, :, np.newaxis] * SMALL_UNCERTAINTY,
                (BICYCLE_ESTIMATION_D_X, BICYCLE_ESTIMATION_D_X, states.count),
            ).copy()

        return NumPyGaussianBelief(mean=augmented, covariance=covariances)

    @cached_property
    def observation_matrix(self) -> ObservationMatrix:
        # NOTE: We observe the pose directly.
        return np.array(
            [
                [1, 0, 0, 0, 0, 0],
                [0, 1, 0, 0, 0, 0],
                [0, 0, 1, 0, 0, 0],
            ]
        )


@dataclass(kw_only=True, frozen=True)
class NumPyBicycleObstacleModel(
    ObstacleModel[
        NumPyBicycleObstacleStatesHistory,
        NumPyBicycleObstacleStates,
        NumPyBicycleObstacleInputs,
        NumPyBicycleObstacleCovariances,
        NumPyBicycleObstacleStateSequences,
    ]
):
    """Propagates bicycle kinematics forward given states and control inputs."""

    model: NumPyBicycleStateEstimationModel
    process_noise_covariance: ProcessNoiseCovarianceArray
    predictor: KalmanFilter

    @staticmethod
    def unscented(
        *,
        time_step_size: float,
        wheelbase: float = 1.0,
        rear_axle_distance: float = 0.0,
        process_noise_covariance: NumPyNoiseCovarianceDescription = 1e-3,
        sigma_point_spread: float = 1.0,
        prior_knowledge: float = 2.0,
    ) -> "NumPyBicycleObstacleModel":
        """Creates a bicycle obstacle model for obstacle state prediction. This
        model uses the unscented transform for propagating state information through
        the nonlinear bicycle dynamics.

        Args:
            time_step_size: Time step size for state propagation.
            wheelbase: The wheelbase length of the bicycle.
            rear_axle_distance: Distance from the reference point to the rear axle.
            process_noise_covariance: The process noise covariance, either as a
                full covariance array, a diagonal covariance vector, or a scalar
                variance representing isotropic noise.
            sigma_point_spread: Spread of sigma points (α) for the unscented
                transform, controlling how far the sigma points are from the mean.
            prior_knowledge: Prior knowledge (β) about the state distribution
                for the unscented transform.
        """
        return NumPyBicycleObstacleModel(
            model=NumPyBicycleStateEstimationModel.create(
                time_step_size=time_step_size,
                wheelbase=wheelbase,
                rear_axle_distance=rear_axle_distance,
            ),
            process_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                process_noise_covariance, dimension=BICYCLE_ESTIMATION_D_X
            ),
            predictor=NumPyUnscentedKalmanFilter.create(
                alpha=sigma_point_spread, beta=prior_knowledge
            ),
        )

    def forward(
        self,
        *,
        states: NumPyBicycleObstacleStates,
        inputs: NumPyBicycleObstacleInputs,
        covariances: NumPyBicycleObstacleCovariances | None,
        horizon: int,
    ) -> NumPyBicycleObstacleStateSequences:
        beliefs = []
        last = self.model.initial_belief_from(
            states=states, inputs=inputs, covariances=covariances
        )

        for _ in range(horizon):
            beliefs.append(
                last := self.predictor.predict(
                    belief=last,
                    state_transition=self.model,
                    process_noise_covariance=self.process_noise_covariance,
                )
            )

        return NumPyBicycleObstacleStateSequences.create(beliefs)


@dataclass(frozen=True)
class NumPyFiniteDifferenceBicycleStateEstimator(
    ObstacleStateEstimator[
        NumPyBicycleObstacleStatesHistory,
        NumPyBicycleObstacleStates,
        NumPyBicycleObstacleInputs,
    ]
):
    time_step_size: float
    wheelbase: float
    rear_axle_distance: float = 0.0

    @staticmethod
    def create(
        *, time_step_size: float, wheelbase: float, rear_axle_distance: float = 0.0
    ) -> "NumPyFiniteDifferenceBicycleStateEstimator":
        return NumPyFiniteDifferenceBicycleStateEstimator(
            time_step_size=time_step_size,
            wheelbase=wheelbase,
            rear_axle_distance=rear_axle_distance,
        )

    def estimate_from(
        self, history: NumPyBicycleObstacleStatesHistory
    ) -> EstimatedObstacleStates[
        NumPyBicycleObstacleStates, NumPyBicycleObstacleInputs, None
    ]:
        """Estimates current states and inputs from pose history using finite differences. Zeros
        are assumed for states that cannot be estimated due to insufficient history length.

        Computes the following quantities from the kinematic bicycle model:

        **Speed** (requires T ≥ 2):
            Magnitude of displacement with sign from heading projection:
            $$v_t = \\text{sign}(v_t) \\cdot \\frac{\\sqrt{(x_t - x_{t-1})^2 + (y_t - y_{t-1})^2}}{\\Delta t}$$

        **Acceleration** (requires T ≥ 3):
            Change in speed over time:
            $$a_t = \\frac{v_t - v_{t-1}}{\\Delta t}$$

        **Steering angle** (requires T ≥ 2):
            Derived from the kinematic bicycle model, where $\\omega = \\frac{d\\theta}{dt}$ is the heading rate:
            $$\\delta_t = \\arctan\\left(\\frac{L \\cdot \\omega_t}{\\sqrt{v_t^2 - l_r^2 \\omega_t^2}}\\right)$$
            Zero when $|v_t| < \\epsilon$ or $v_t^2 \\leq l_r^2 \\omega_t^2$ (estimate becomes unreliable).

        Args:
            history: Obstacle pose history containing at least one entry.
        """
        assert history.horizon > 0, "History must contain at least one state."

        filter_invalid_short = invalid_obstacle_filter_from(history, check_recent=2)
        filter_invalid_long = invalid_obstacle_filter_from(history, check_recent=3)

        speeds = self.estimate_speeds_from(history)
        accelerations = self.estimate_accelerations_from(history, speeds=speeds)
        steering_angles = self.estimate_steering_angles_from(history, speeds=speeds)

        return EstimatedObstacleStates(
            states=NumPyBicycleObstacleStates.create(
                x=history.x()[-1],
                y=history.y()[-1],
                heading=history.heading()[-1],
                speed=filter_invalid_short(speeds),
            ),
            inputs=NumPyBicycleObstacleInputs.create(
                accelerations=filter_invalid_long(accelerations),
                steering_angles=filter_invalid_short(steering_angles),
            ),
            covariance=None,
        )

    def invalid_obstacle_mask_from(
        self, history: NumPyBicycleObstacleStatesHistory
    ) -> Float[Array, " K"]:
        return np.any(
            np.isnan(history.x()[-3:])
            | np.isnan(history.y()[-3:])
            | np.isnan(history.heading()[-3:]),
            axis=0,
        )

    def estimate_speeds_from(
        self, history: NumPyBicycleObstacleStatesHistory
    ) -> Float[Array, " K"]:
        if history.horizon < 2:
            return np.zeros((history.count,))

        x = history.x()
        y = history.y()
        heading = history.heading()

        return self._estimate_speeds_from(
            x_current=x[-1],
            y_current=y[-1],
            x_previous=x[-2],
            y_previous=y[-2],
            heading_current=heading[-1],
        )

    def estimate_accelerations_from(
        self,
        history: NumPyBicycleObstacleStatesHistory,
        *,
        speeds: Float[Array, " K"],
    ) -> Float[Array, " K"]:
        if history.horizon < 3:
            return np.zeros((history.count,))

        x = history.x()
        y = history.y()
        heading = history.heading()

        return self._estimate_accelerations_from(
            speeds_current=speeds,
            speeds_previous=self._estimate_speeds_from(
                x_current=x[-2],
                y_current=y[-2],
                x_previous=x[-3],
                y_previous=y[-3],
                heading_current=heading[-2],
            ),
        )

    def estimate_steering_angles_from(
        self, history: NumPyBicycleObstacleStatesHistory, *, speeds: Float[Array, " K"]
    ) -> Float[Array, " K"]:
        if history.horizon < 2:
            return np.zeros((history.count,))

        heading = history.heading()

        return self._estimate_steering_angles_from(
            heading_current=heading[-1], heading_previous=heading[-2], speeds=speeds
        )

    def _estimate_speeds_from(
        self,
        *,
        x_current: Float[Array, " K"],
        y_current: Float[Array, " K"],
        x_previous: Float[Array, " K"],
        y_previous: Float[Array, " K"],
        heading_current: Float[Array, " K"],
    ) -> Float[Array, " K"]:
        delta_x = x_current - x_previous
        delta_y = y_current - y_previous

        projection = delta_x * np.cos(heading_current) + delta_y * np.sin(
            heading_current
        )
        magnitude = np.sqrt(delta_x**2 + delta_y**2)
        speeds = np.sign(projection) * magnitude / self.time_step_size

        assert speeds.shape == (x_current.shape[0],)

        return speeds

    def _estimate_accelerations_from(
        self, *, speeds_current: Float[Array, " K"], speeds_previous: Float[Array, " K"]
    ) -> Float[Array, " K"]:
        return (speeds_current - speeds_previous) / self.time_step_size

    def _estimate_steering_angles_from(
        self,
        *,
        heading_current: Float[Array, " K"],
        heading_previous: Float[Array, " K"],
        speeds: Float[Array, " K"],
    ) -> Float[Array, " K"]:
        heading_velocity = (heading_current - heading_previous) / self.time_step_size
        l_r = self.rear_axle_distance

        with np.errstate(invalid="ignore"):
            denominator_squared = speeds**2 - l_r**2 * heading_velocity**2
            is_reliable = (np.abs(speeds) > 1e-6) & (denominator_squared > 0)

            steering_angles = np.where(
                is_reliable,
                np.arctan(
                    heading_velocity
                    * self.wheelbase
                    * np.sign(speeds)
                    / np.sqrt(np.maximum(denominator_squared, 1e-12))
                ),
                0.0,
            )

        return steering_angles


@dataclass(frozen=True)
class NumPyKfBicycleStateEstimator(
    ObstacleStateEstimator[
        NumPyBicycleObstacleStatesHistory,
        NumPyBicycleObstacleStates,
        NumPyBicycleObstacleInputs,
        NumPyBicycleObstacleCovariances,
    ]
):
    """Kalman Filter state estimator for bicycle model obstacles."""

    process_noise_covariance: ProcessNoiseCovarianceArray
    observation_noise_covariance: ObservationNoiseCovarianceArray
    model: NumPyBicycleStateEstimationModel
    estimator: KalmanFilter
    noise_model: NumPyNoiseModelProvider | None = None

    @staticmethod
    def ekf(
        *,
        time_step_size: float,
        wheelbase: float,
        rear_axle_distance: float = 0.0,
        process_noise_covariance: NumPyNoiseCovarianceDescription,
        observation_noise_covariance: NumPyNoiseCovarianceDescription,
        initial_state_covariance: EstimationStateCovarianceArray | None = None,
        noise_model: NumPyNoiseModelProvider | None = None,
    ) -> "NumPyKfBicycleStateEstimator":
        """Creates an EKF state estimator for the bicycle model with the specified noise
        covariances.

        Args:
            time_step_size: The time step size for the state transition model.
            wheelbase: The wheelbase of the bicycle.
            rear_axle_distance: Distance from the reference point to the rear axle.
            process_noise_covariance: The process noise covariance, either as a full
                matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            observation_noise_covariance: The observation noise covariance, either as a
                full matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            initial_state_covariance: The initial state covariance for the Kalman filter.
                If not provided, low uncertainty will be assumed for observed states and high
                uncertainty for speed and inputs.
            noise_model: Optional noise model provider for adaptive noise estimation.
        """
        return NumPyKfBicycleStateEstimator(
            process_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                process_noise_covariance, dimension=BICYCLE_ESTIMATION_D_X
            ),
            observation_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                observation_noise_covariance, dimension=BICYCLE_OBSERVATION_D_O
            ),
            model=NumPyBicycleStateEstimationModel.create(
                time_step_size=time_step_size,
                wheelbase=wheelbase,
                rear_axle_distance=rear_axle_distance,
                initial_state_covariance=initial_state_covariance,
            ),
            estimator=NumPyExtendedKalmanFilter.create(noise_model=noise_model),
        )

    @staticmethod
    def ukf(
        *,
        time_step_size: float,
        wheelbase: float,
        rear_axle_distance: float = 0.0,
        process_noise_covariance: NumPyNoiseCovarianceDescription,
        observation_noise_covariance: NumPyNoiseCovarianceDescription,
        initial_state_covariance: EstimationStateCovarianceArray | None = None,
        noise_model: NumPyNoiseModelProvider | None = None,
    ) -> "NumPyKfBicycleStateEstimator":
        """Creates a UKF state estimator for the bicycle model with the specified noise
        covariances.

        Args:
            time_step_size: The time step size for the state transition model.
            wheelbase: The wheelbase of the bicycle.
            rear_axle_distance: Distance from the reference point to the rear axle.
            process_noise_covariance: The process noise covariance, either as a full
                matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            observation_noise_covariance: The observation noise covariance, either as a
                full matrix, a vector of diagonal entries, or a scalar for isotropic noise.
            initial_state_covariance: The initial state covariance for the Kalman filter.
                If not provided, low uncertainty will be assumed for observed states and high
                uncertainty for speed and inputs.
            noise_model: Optional noise model provider for adaptive noise estimation.
        """
        return NumPyKfBicycleStateEstimator(
            process_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                process_noise_covariance, dimension=BICYCLE_ESTIMATION_D_X
            ),
            observation_noise_covariance=numpy_kalman_filter.standardize_noise_covariance(
                observation_noise_covariance, dimension=BICYCLE_OBSERVATION_D_O
            ),
            model=NumPyBicycleStateEstimationModel.create(
                time_step_size=time_step_size,
                wheelbase=wheelbase,
                rear_axle_distance=rear_axle_distance,
                initial_state_covariance=initial_state_covariance,
            ),
            estimator=NumPyUnscentedKalmanFilter.create(noise_model=noise_model),
        )

    def estimate_from(
        self, history: NumPyBicycleObstacleStatesHistory
    ) -> EstimatedObstacleStates[
        NumPyBicycleObstacleStates,
        NumPyBicycleObstacleInputs,
        NumPyBicycleObstacleCovariances,
    ]:
        estimate = cast(
            BicycleGaussianBelief,
            self.estimator.filter(
                self.model.observations_from(history),
                initial_state_covariance=self.model.initial_state_covariance,
                state_transition=self.model,
                process_noise_covariance=self.process_noise_covariance,
                observation_noise_covariance=self.observation_noise_covariance,
                observation_matrix=self.model.observation_matrix,
            ),
        )

        return EstimatedObstacleStates(
            states=self.model.states_from(estimate),
            inputs=self.model.inputs_from(estimate),
            covariance=estimate.covariance,
        )


def simulate(
    inputs: ControlInputBatchArray,
    initial: StatesAtTimeStep,
    *,
    time_step_size: float,
    wheelbase: float,
    rear_axle_distance: float = 0.0,
    speed_limits: tuple[float, float],
    steering_limits: tuple[float, float],
    acceleration_limits: tuple[float, float],
) -> StateBatchArray:
    horizon = inputs.shape[0]
    rollout_count = inputs.shape[2]
    states = np.zeros((horizon, BICYCLE_D_X, rollout_count))
    current = initial

    for t in range(horizon):
        current = step(
            current,
            inputs[t],
            time_step_size=time_step_size,
            wheelbase=wheelbase,
            rear_axle_distance=rear_axle_distance,
            speed_limits=speed_limits,
            steering_limits=steering_limits,
            acceleration_limits=acceleration_limits,
        )
        states[t] = current

    return states


def step(
    state: StatesAtTimeStep,
    control: ControlInputsAtTimeStep,
    *,
    time_step_size: float,
    wheelbase: float,
    rear_axle_distance: float = 0.0,
    speed_limits: tuple[float, float],
    steering_limits: tuple[float, float],
    acceleration_limits: tuple[float, float],
) -> StatesAtTimeStep:
    x, y, theta, v = state[0], state[1], state[2], state[3]
    a, delta = control[0], control[1]
    acceleration = np.clip(a, *acceleration_limits)
    steering = np.clip(delta, *steering_limits)

    beta = np.arctan(rear_axle_distance / wheelbase * np.tan(steering))

    new_x = x + v * np.cos(theta + beta) * time_step_size
    new_y = y + v * np.sin(theta + beta) * time_step_size
    new_theta = theta + v * np.cos(beta) * np.tan(steering) / wheelbase * time_step_size
    new_v = np.clip(v + acceleration * time_step_size, *speed_limits)

    return np.stack([new_x, new_y, new_theta, new_v])



from dataclasses import dataclass

from faran.types import (
    jaxtyped,
    CostFunction,
    ControlInputBatch,
    Trajectory,
    BoundaryPoints,
    BoundaryWidthsDescription,
    JaxReferencePoints,
    JaxBoundaryDistance,
    JaxBoundaryDistanceExtractor,
    JaxCosts,
    JaxPathParameters,
    JaxPositions,
    JaxLateralPositions,
    JaxPositionExtractor,
)
from faran.states import JaxSimpleCosts

from jaxtyping import Float, Array as JaxArray, Scalar

import jax
import jax.numpy as jnp
import numpy as np


@dataclass(frozen=True)
class JaxBoundaryCost[StateT](CostFunction[ControlInputBatch, StateT, JaxCosts]):
    """Penalizes states that approach within a threshold distance of the corridor edges."""

    distance: JaxBoundaryDistanceExtractor[StateT, JaxBoundaryDistance]
    distance_threshold: Scalar
    weight: Scalar

    @staticmethod
    def create[S](
        *,
        distance: JaxBoundaryDistanceExtractor[S, JaxBoundaryDistance],
        distance_threshold: float,
        weight: float,
    ) -> "JaxBoundaryCost[S]":
        return JaxBoundaryCost(
            distance=distance,
            distance_threshold=jnp.array(distance_threshold),
            weight=jnp.array(weight),
        )

    def __call__(self, *, inputs: ControlInputBatch, states: StateT) -> JaxCosts:
        return JaxSimpleCosts(
            boundary_cost(
                distance=self.distance(states=states).array,
                distance_threshold=self.distance_threshold,
                weight=self.weight,
            )
        )


@dataclass(kw_only=True, frozen=True)
class JaxFixedWidthBoundary[StateT](
    JaxBoundaryDistanceExtractor[StateT, JaxBoundaryDistance]
):
    """Fixed-width corridor boundary around a reference trajectory."""

    reference: Trajectory[
        JaxPathParameters, JaxReferencePoints, JaxPositions, JaxLateralPositions
    ]
    position_extractor: JaxPositionExtractor[StateT]
    _left: Scalar
    _right: Scalar

    @staticmethod
    def create[S](
        *,
        reference: Trajectory[
            JaxPathParameters, JaxReferencePoints, JaxPositions, JaxLateralPositions
        ],
        position_extractor: JaxPositionExtractor[S],
        left: float,
        right: float,
    ) -> "JaxFixedWidthBoundary[S]":
        """Creates a fixed-width boundary distance extractor.

        This component assumes a fixed-width corridor around a reference trajectory. The left and
        right widths can be different (asymmetric corridor).

        Args:
            reference: The reference trajectory defining the center of the corridor.
            position_extractor: Function to extract positions from states.
            left: The width of the left side of the corridor.
            right: The width of the right side of the corridor.
        """
        assert left >= -right, (
            f"The boundaries appear to be inverted. Left: {left}, Right: {right}. "
            f"Make sure the total width (left + right) is non-negative, got {left + right}."
        )

        return JaxFixedWidthBoundary(
            reference=reference,
            position_extractor=position_extractor,
            _left=jnp.array(left),
            _right=jnp.array(right),
        )

    def __call__(self, *, states: StateT) -> JaxBoundaryDistance:
        positions = self.position_extractor(states)
        lateral = self.reference.lateral(positions)

        return JaxBoundaryDistance(
            boundary_distance(lateral=lateral.array, left=self._left, right=self._right)
        )

    def left(self, *, sample_count: int = 100) -> BoundaryPoints:
        s = JaxPathParameters(
            jnp.linspace(0, self.reference.path_length, sample_count).reshape(-1, 1)
        )

        return np.asarray(
            (
                self.reference.query(s).positions_array
                - self._left * self.reference.normal(s).array
            )[..., 0]
        )

    def right(self, *, sample_count: int = 100) -> BoundaryPoints:
        s = JaxPathParameters(
            jnp.linspace(0, self.reference.path_length, sample_count).reshape(-1, 1)
        )

        return np.asarray(
            (
                self.reference.query(s).positions_array
                + self._right * self.reference.normal(s).array
            )[..., 0]
        )


@dataclass(kw_only=True, frozen=True)
class JaxPiecewiseFixedWidthBoundary[StateT](
    JaxBoundaryDistanceExtractor[StateT, JaxBoundaryDistance]
):
    """Piecewise fixed-width corridor boundary with segment-varying widths."""

    reference: Trajectory[
        JaxPathParameters, JaxReferencePoints, JaxPositions, JaxLateralPositions
    ]
    position_extractor: JaxPositionExtractor[StateT]
    breakpoints: Float[JaxArray, " B"]
    left_widths: Float[JaxArray, " B"]
    right_widths: Float[JaxArray, " B"]

    @staticmethod
    def create[S](
        *,
        reference: Trajectory[
            JaxPathParameters, JaxReferencePoints, JaxPositions, JaxLateralPositions
        ],
        position_extractor: JaxPositionExtractor[S],
        widths: BoundaryWidthsDescription,
    ) -> "JaxPiecewiseFixedWidthBoundary[S]":
        """Creates a piecewise fixed-width boundary distance extractor.

        This component assumes a piecewise constant corridor width around a reference trajectory.
        Different segments of the trajectory can have different left and right widths.

        Args:
            reference: The reference trajectory defining the center of the corridor.
            position_extractor: Function to extract positions from states.
            widths: A mapping from longitudinal breakpoints to {"left": float, "right": float}
                   dictionaries defining the corridor widths for each segment.
        """
        sorted_breakpoints = sorted(widths.keys())

        return JaxPiecewiseFixedWidthBoundary(
            reference=reference,
            position_extractor=position_extractor,
            breakpoints=jnp.array(sorted_breakpoints),
            left_widths=jnp.array([widths[s]["left"] for s in sorted_breakpoints]),
            right_widths=jnp.array([widths[s]["right"] for s in sorted_breakpoints]),
        )

    def __call__(self, *, states: StateT) -> JaxBoundaryDistance:
        positions = self.position_extractor(states)
        lateral = self.reference.lateral(positions)
        longitudinal = self.reference.longitudinal(positions)

        return JaxBoundaryDistance(
            piecewise_boundary_distance(
                lateral=lateral.array,
                longitudinal=longitudinal.array,
                breakpoints=self.breakpoints,
                left_widths=self.left_widths,
                right_widths=self.right_widths,
            )
        )

    def left(self, *, sample_count: int = 100) -> BoundaryPoints:
        s_values = jnp.linspace(0, self.reference.path_length, sample_count)
        s = JaxPathParameters(s_values.reshape(-1, 1))

        return np.asarray(
            piecewise_left_boundary_points(
                positions=self.reference.query(s).positions_array,
                normals=self.reference.normal(s).array,
                s_values=s_values,
                breakpoints=self.breakpoints,
                left_widths=self.left_widths,
            )
        )

    def right(self, *, sample_count: int = 100) -> BoundaryPoints:
        s_values = jnp.linspace(0, self.reference.path_length, sample_count)
        s = JaxPathParameters(s_values.reshape(-1, 1))

        return np.asarray(
            piecewise_right_boundary_points(
                positions=self.reference.query(s).positions_array,
                normals=self.reference.normal(s).array,
                s_values=s_values,
                breakpoints=self.breakpoints,
                right_widths=self.right_widths,
            )
        )


@jax.jit
@jaxtyped
def boundary_cost(
    *, distance: Float[JaxArray, "T M"], distance_threshold: Scalar, weight: Scalar
) -> Float[JaxArray, "T M"]:
    cost = distance_threshold - distance
    return weight * jnp.clip(cost, 0, None)


@jax.jit
@jaxtyped
def boundary_distance(
    *, lateral: Float[JaxArray, "T M"], left: Scalar, right: Scalar
) -> Float[JaxArray, "T M"]:
    distance_to_left = left + lateral
    distance_to_right = right - lateral
    return jnp.minimum(distance_to_left, distance_to_right)


@jax.jit
@jaxtyped
def piecewise_boundary_distance(
    *,
    lateral: Float[JaxArray, "T M"],
    longitudinal: Float[JaxArray, "T M"],
    breakpoints: Float[JaxArray, " N"],
    left_widths: Float[JaxArray, " N"],
    right_widths: Float[JaxArray, " N"],
) -> Float[JaxArray, "T M"]:
    segment_indices = jnp.searchsorted(breakpoints, longitudinal, side="right") - 1
    segment_indices = jnp.clip(segment_indices, 0, len(breakpoints) - 1)

    left = left_widths[segment_indices]
    right = right_widths[segment_indices]

    distance_to_left = left + lateral
    distance_to_right = right - lateral

    return jnp.minimum(distance_to_left, distance_to_right)


@jax.jit
@jaxtyped
def piecewise_left_boundary_points(
    *,
    positions: Float[JaxArray, "N 2 1"],
    normals: Float[JaxArray, "N 2 1"],
    s_values: Float[JaxArray, " N"],
    breakpoints: Float[JaxArray, " K"],
    left_widths: Float[JaxArray, " K"],
) -> Float[JaxArray, "N 2"]:
    segment_indices = jnp.searchsorted(breakpoints, s_values, side="right") - 1
    segment_indices = jnp.clip(segment_indices, 0, len(breakpoints) - 1)
    left = left_widths[segment_indices]

    return (positions - left[:, jnp.newaxis, jnp.newaxis] * normals)[..., 0]


@jax.jit
@jaxtyped
def piecewise_right_boundary_points(
    *,
    positions: Float[JaxArray, "N 2 1"],
    normals: Float[JaxArray, "N 2 1"],
    s_values: Float[JaxArray, " N"],
    breakpoints: Float[JaxArray, " K"],
    right_widths: Float[JaxArray, " K"],
) -> Float[JaxArray, "N 2"]:
    segment_indices = jnp.searchsorted(breakpoints, s_values, side="right") - 1
    segment_indices = jnp.clip(segment_indices, 0, len(breakpoints) - 1)
    right = right_widths[segment_indices]

    return (positions + right[:, jnp.newaxis, jnp.newaxis] * normals)[..., 0]



from typing import Protocol

from faran.types.array import Array, DataType

from jaxtyping import Float


class Error(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T M"]:
        """Returns the error as a NumPy array."""
        ...


class PositionExtractor[StateBatchT, PositionsT](Protocol):
    def __call__(self, states: StateBatchT, /) -> PositionsT:
        """Extracts (x, y) positions from a batch of states."""
        ...


class ContouringCost[InputBatchT, StateBatchT, ErrorT = Error](Protocol):
    def error(self, *, states: StateBatchT) -> ErrorT:
        """Computes the contouring error for the given states."""
        ...


class LagCost[InputBatchT, StateBatchT, ErrorT = Error](Protocol):
    def error(self, *, states: StateBatchT) -> ErrorT:
        """Computes the lag error for the given states."""
        ...



from typing import Final, TypeAlias

from faran.namespace import types as _types, classes as _classes
from faran.factory import (
    mppi as _mppi,
    update as _update,
    padding as _padding,
    filters as _filters,
)
from faran.costs import (
    costs as _costs,
    distance as _distance,
    boundary as _boundary,
    risk as _risk,
)
from faran.models import model as _model
from faran.obstacles import obstacles as _obstacles
from faran.predictors import predictor as _predictor
from faran.samplers import sampler as _sampler
from faran.trajectories import trajectory as _trajectory
from faran.states import extract as _extract
from faran.filters import noise as _noise

types: TypeAlias = _types.jax
classes: Final = _classes.jax
mppi: Final = _mppi.jax
update: Final = _update.jax
padding: Final = _padding.jax
filters: Final = _filters.jax
costs: Final = _costs.jax
distance: Final = _distance.jax
boundary: Final = _boundary.jax
risk: Final = _risk.jax
model: Final = _model.jax
obstacles: Final = _obstacles.jax
predictor: Final = _predictor.jax
sampler: Final = _sampler.jax
trajectory: Final = _trajectory.jax
noise: Final = _noise.jax
extract: Final = _extract



from typing import Protocol, Final

from faran.types.array import Array, DataType

from jaxtyping import Float
from numtypes import D

POSE_D_O: Final = 3

type PoseD_o = D[3]
"""Dimension of a single obstacle pose state (x, y, heading)."""


class SampledObstacleStates(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_o K N"]:
        """Returns the sampled states of obstacles as a NumPy array."""
        ...

    @property
    def horizon(self) -> int:
        """The time horizon over which the obstacle states are defined."""
        ...

    @property
    def dimension(self) -> int:
        """The dimension of a single obstacle state."""
        ...

    @property
    def count(self) -> int:
        """The number of obstacles."""
        ...

    @property
    def sample_count(self) -> int:
        """The number of samples per obstacle."""
        ...


class SampledObstaclePositions(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T 2 K N"]:
        """Returns the sampled positions of obstacles as a NumPy array."""
        ...

    def x(self) -> Float[Array, "T K N"]:
        """Returns the x positions of obstacles over time and samples."""
        ...

    def y(self) -> Float[Array, "T K N"]:
        """Returns the y positions of obstacles over time and samples."""
        ...

    @property
    def horizon(self) -> int:
        """The time horizon over which the positions are defined."""
        ...

    @property
    def count(self) -> int:
        """The number of obstacles."""
        ...

    @property
    def sample_count(self) -> int:
        """The number of samples per obstacle."""
        ...


class SampledObstacleHeadings(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T K N"]:
        """Returns the sampled headings of obstacles as a NumPy array."""
        ...

    def heading(self) -> Float[Array, "T K N"]:
        """Returns the headings of obstacles over time and samples."""
        ...

    @property
    def horizon(self) -> int:
        """The time horizon over which the headings are defined."""
        ...

    @property
    def count(self) -> int:
        """The number of obstacles."""
        ...

    @property
    def sample_count(self) -> int:
        """The number of samples per obstacle."""
        ...


class ObstacleStatesForTimeStep[ObstacleStatesT](Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "D_o K"]:
        """Returns the mean states of obstacles at a single time step as a NumPy array."""
        ...

    def replicate(self, *, horizon: int) -> ObstacleStatesT:
        """Replicates the obstacle states over the given time horizon."""
        ...

    @property
    def dimension(self) -> int:
        """The dimension of a single obstacle state."""
        ...

    @property
    def count(self) -> int:
        """The number of obstacles."""
        ...


class ObstacleStates[SingleSampleT](Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T D_o K"]:
        """Returns the mean states of obstacles as a NumPy array."""
        ...

    def covariance(self) -> Float[Array, "T D_o D_o K"] | None:
        """Returns the covariance matrices for the obstacle states x, y, heading over time, if
        one exists."""
        ...

    def single(self) -> SingleSampleT:
        """Returns the single (mean) sample of the obstacle states."""
        ...

    @property
    def horizon(self) -> int:
        """The time horizon over which the obstacle states are defined."""
        ...

    @property
    def dimension(self) -> int:
        """The dimension of a single obstacle state."""
        ...

    @property
    def count(self) -> int:
        """The number of obstacles."""
        ...


class ObstacleStateProvider[ObstacleStatesT](Protocol):
    def __call__(self) -> ObstacleStatesT:
        """Provides a forecast of states over the time horizon for each obstacle."""
        ...


class ObstacleStateSampler[ObstacleStatesT, SampledObstacleStatesT](Protocol):
    def __call__(
        self, states: ObstacleStatesT, *, count: int
    ) -> SampledObstacleStatesT:
        """Samples the obstacle state forecasts."""
        ...


class Distance(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T V M N"]:
        """Returns the distances between ego parts and obstacles as a NumPy array."""
        ...

    @property
    def horizon(self) -> int:
        """The time horizon over which the distances are computed."""
        ...

    @property
    def vehicle_parts(self) -> int:
        """The number of ego vehicle parts for which distances are computed."""
        ...

    @property
    def rollout_count(self) -> int:
        """The number of rollouts for which distances are computed."""
        ...

    @property
    def sample_count(self) -> int:
        """The number of obstacle samples for which distances are computed."""
        ...


class SampledObstaclePositionExtractor[SampledStatesT, PositionsT](Protocol):
    def __call__(self, states: SampledStatesT, /) -> PositionsT:
        """Extracts the positions from the given sampled obstacle states."""
        ...


class SampledObstacleHeadingExtractor[SampledStatesT, HeadingsT](Protocol):
    def __call__(self, states: SampledStatesT, /) -> HeadingsT:
        """Extracts the headings from the given sampled obstacle states."""
        ...


class DistanceExtractor[StateBatchT, SampledObstacleStatesT, DistanceT](Protocol):
    def __call__(
        self, *, states: StateBatchT, obstacle_states: SampledObstacleStatesT
    ) -> DistanceT:
        """Computes the distances between each part of the ego and the corresponding closest
        obstacles."""
        ...


class SampleCostFunction[StateBatchT, SampledObstacleStatesT, CostsT](Protocol):
    def __call__(
        self, *, states: StateBatchT, samples: SampledObstacleStatesT
    ) -> CostsT:
        """Computes the cost given the states and sampled obstacle states."""
        ...


class Risk(Protocol):
    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T M"]:
        """Returns the risk values as a NumPy array."""
        ...

    @property
    def horizon(self) -> int:
        """The time horizon over which the risk is computed."""
        ...

    @property
    def rollout_count(self) -> int:
        """The number of rollouts for which the risk is computed."""
        ...


class RiskMetric[CostFunctionT, StateBatchT, ObstacleStatesT, SamplerT, RiskT](
    Protocol
):
    def compute(
        self,
        cost_function: CostFunctionT,
        *,
        states: StateBatchT,
        obstacle_states: ObstacleStatesT,
        sampler: SamplerT,
    ) -> RiskT:
        """Computes the risk metric based on the provided cost function."""
        ...

    @property
    def name(self) -> str:
        """Returns a human-readable name for the risk metric."""
        ...



from typing import Protocol
from dataclasses import dataclass
from functools import cached_property

from faran.types.array import Array, jaxtyped
from faran.types.trajectories.common import (
    D_R,
    PathParameters,
    ReferencePoints,
    Positions,
    LateralPositions,
    LongitudinalPositions,
    Normals,
)

from jaxtyping import Float

import numpy as np


@jaxtyped
@dataclass(frozen=True)
class NumPyPathParameters(PathParameters):
    array: Float[Array, "T M"]

    def __array__(self) -> Float[Array, "T M"]:
        return self.array

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self.array.shape[1]


class NumPyPositions(Positions, Protocol):
    @property
    def array(self) -> Float[Array, "T 2 M"]:
        """Returns the positions as a NumPy array."""
        ...


@jaxtyped
@dataclass(frozen=True)
class NumPySimplePositions(NumPyPositions):
    _x: Float[Array, "T M"]
    _y: Float[Array, "T M"]

    @staticmethod
    def create(
        *, x: Float[Array, "T M"], y: Float[Array, "T M"]
    ) -> "NumPySimplePositions":
        """Creates a NumPy positions instance from x and y coordinate arrays."""
        return NumPySimplePositions(_x=x, _y=y)

    def __array__(self) -> Float[Array, "T 2 M"]:
        return self.array

    def x(self) -> Float[Array, "T M"]:
        return self._x

    def y(self) -> Float[Array, "T M"]:
        return self._y

    @property
    def horizon(self) -> int:
        return self._x.shape[0]

    @property
    def rollout_count(self) -> int:
        return self._x.shape[1]

    @property
    def array(self) -> Float[Array, "T 2 M"]:
        return self._array

    @cached_property
    def _array(self) -> Float[Array, "T 2 M"]:
        return np.stack([self._x, self._y], axis=1)


@jaxtyped
@dataclass(frozen=True)
class NumPyHeadings:
    _heading: Float[Array, "T M"]

    @staticmethod
    def create(
        *,
        heading: Float[Array, "T M"],
    ) -> "NumPyHeadings":
        """Creates a NumPy headings instance from an array of headings."""
        return NumPyHeadings(heading)

    def __array__(self) -> Float[Array, "T M"]:
        return self.array

    def heading(self) -> Float[Array, "T M"]:
        return self._heading

    @property
    def horizon(self) -> int:
        return self._heading.shape[0]

    @property
    def rollout_count(self) -> int:
        return self._heading.shape[1]

    @property
    def array(self) -> Float[Array, "T M"]:
        return self._heading


@jaxtyped
@dataclass(frozen=True)
class NumPyReferencePoints(ReferencePoints):
    array: Float[Array, f"T {D_R} M"]

    @staticmethod
    def create(
        *, x: Float[Array, "T M"], y: Float[Array, "T M"], heading: Float[Array, "T M"]
    ) -> "NumPyReferencePoints":
        """Creates a NumPy reference points instance from x, y, and heading arrays."""
        return NumPyReferencePoints(array=np.stack([x, y, heading], axis=1))

    def __array__(self) -> Float[Array, f"T {D_R} M"]:
        return self.array

    def x(self) -> Float[Array, "T M"]:
        return self.array[:, 0]

    def y(self) -> Float[Array, "T M"]:
        return self.array[:, 1]

    def heading(self) -> Float[Array, "T M"]:
        return self.array[:, 2]

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def positions(self) -> Float[Array, "T 2 M"]:
        return self.array[:, :2]


@jaxtyped
@dataclass(frozen=True)
class NumPyLateralPositions(LateralPositions):
    _array: Float[Array, "T M"]

    @staticmethod
    def create(array: Float[Array, "T M"]) -> "NumPyLateralPositions":
        """Creates a NumPy lateral positions instance from an array."""
        return NumPyLateralPositions(array)

    def __array__(self) -> Float[Array, "T M"]:
        return self.array

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self.array.shape[1]

    @property
    def array(self) -> Float[Array, "T M"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class NumPyLongitudinalPositions(LongitudinalPositions):
    _array: Float[Array, "T M"]

    @staticmethod
    def create(array: Float[Array, "T M"]) -> "NumPyLongitudinalPositions":
        """Creates a NumPy longitudinal positions instance from an array."""
        return NumPyLongitudinalPositions(array)

    def __array__(self) -> Float[Array, "T M"]:
        return self.array

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self.array.shape[1]

    @property
    def array(self) -> Float[Array, "T M"]:
        return self._array


@jaxtyped
@dataclass(frozen=True)
class NumPyNormals(Normals):
    _array: Float[Array, "T 2 M"]

    @staticmethod
    def create(*, x: Float[Array, "T M"], y: Float[Array, "T M"]) -> "NumPyNormals":
        """Creates a NumPy normals instance from x and y coordinate arrays."""
        return NumPyNormals(np.stack([x, y], axis=1))

    def __array__(self) -> Float[Array, "T 2 M"]:
        return self.array

    def x(self) -> Float[Array, "T M"]:
        return self.array[:, 0]

    def y(self) -> Float[Array, "T M"]:
        return self.array[:, 1]

    @property
    def horizon(self) -> int:
        return self.array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self.array.shape[2]

    @property
    def array(self) -> Float[Array, "T 2 M"]:
        return self._array



from typing import Any, NamedTuple


class IdentityNoiseModel(NamedTuple):
    @property
    def state(self) -> None:
        return None

    def __call__[NoiseT](
        self, *, noise: NoiseT, prediction: Any, observation: Any, state: None = None
    ) -> tuple[NoiseT, None]:
        return noise, state


class IdentityNoiseModelProvider(NamedTuple):
    def __call__(
        self, *, obstacle_count: int, observation_matrix: Any, noise: Any
    ) -> IdentityNoiseModel:
        return IdentityNoiseModel()



from typing import Final
from dataclasses import dataclass

from faran.types import (
    jaxtyped,
    JaxControlInputBatchCreator,
    JaxControlInputSequence,
    JaxControlInputBatch,
    JaxSampler,
)
from faran.samplers.accelerated import StandardDeviationDescription, standardize

from jaxtyping import Array as JaxArray, Float, Int, Scalar

import jax
import jax.numpy as jnp
import jax.scipy.special as jspecial

type IntScalar = Int[JaxArray, ""]
type RadicalInverseCarry = tuple[
    IntScalar,  # current_index
    Scalar,  # result
    Scalar,  # factor
]

# fmt: off
PRIMES: Final = jnp.array([
    2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 
    43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97, 
    101, 103, 107, 109, 113, 127, 131, 137, 139, 149, 
    151, 157, 163, 167, 173, 179, 181, 191, 193, 197, 
    199, 211, 223, 227, 229, 233, 239, 241, 251, 257, 
    263, 269, 271, 277, 281, 283, 293, 307, 311, 313, 
    317, 331, 337, 347, 349, 353, 359, 367, 373, 379, 
    383, 389, 397, 401, 409, 419, 421, 431, 433, 439, 
    443, 449, 457, 461, 463, 467, 479, 487, 491, 499, 
    503, 509, 521, 523, 541, 547, 557, 563, 569, 571, 
    577, 587, 593, 599, 601, 607, 613, 617, 619, 631, 
    641, 643, 647, 653, 659, 661, 673, 677, 683, 691, 
    701, 709, 719, 727, 733, 739, 743, 751, 757, 761, 
    769, 773, 787, 797, 809, 811, 821, 823, 827, 829, 
    839, 853, 857, 859, 863, 877, 881, 883, 887, 907, 
    911, 919, 929, 937, 941, 947, 953, 967, 971, 977, 
    983, 991, 997, 1009, 1013, 1019, 1021, 1031, 1033, 
    1039, 1049, 1051, 1061, 1063
])
# fmt: on

MAX_HALTON_DIMENSION: Final = PRIMES.shape[0]
MAX_RADICAL_INVERSE_ITERATIONS: Final = 32


@dataclass(kw_only=True)
class JaxHaltonSplineSampler[
    BatchT: JaxControlInputBatch,
](JaxSampler[JaxControlInputSequence, BatchT]):
    """Perturbs a nominal control sequence using Halton sequences interpolated through cubic splines."""

    standard_deviation: Final[Float[JaxArray, " D_u"]]
    to_batch: Final[JaxControlInputBatchCreator[BatchT]]
    knot_count: Final[int]
    halton_start_index: IntScalar

    _rollout_count: Final[int]

    @staticmethod
    def create[B: JaxControlInputBatch](
        *,
        standard_deviation: StandardDeviationDescription,
        rollout_count: int,
        knot_count: int,
        to_batch: JaxControlInputBatchCreator,
        seed: int,
    ) -> "JaxHaltonSplineSampler":
        return JaxHaltonSplineSampler(
            standard_deviation=standardize.std(standard_deviation),
            to_batch=to_batch,
            knot_count=knot_count,
            halton_start_index=jnp.array(seed),
            _rollout_count=rollout_count,
        )

    def __post_init__(self) -> None:
        assert self.knot_count >= 2, "Knot count must be at least 2."
        assert (
            halton_dimensions_for(
                control_dimension=self.dimension, knot_count=self.knot_count
            )
            <= MAX_HALTON_DIMENSION
        ), (
            f"Halton sequence dimensions (control_dimension * knot_count) cannot exceed {MAX_HALTON_DIMENSION}. "
            f"Got control_dimension={self.dimension} and knot_count={self.knot_count}."
        )

    def sample(self, *, around: JaxControlInputSequence) -> BatchT:
        assert self.knot_count <= around.horizon, (
            f"Knot count ({self.knot_count}) cannot exceed time horizon ({around.horizon})."
        )

        samples, new_halton_index = sample_halton(
            around=around.array,
            standard_deviation=self.standard_deviation,
            rollout_count=self.rollout_count,
            knot_count=self.knot_count,
            halton_start_index=self.halton_start_index,
        )
        self.halton_start_index = new_halton_index
        return self.to_batch(array=samples)

    @property
    def dimension(self) -> int:
        return self.standard_deviation.shape[0]

    @property
    def rollout_count(self) -> int:
        return self._rollout_count


def halton_dimensions_for(*, control_dimension: int, knot_count: int) -> int:
    return control_dimension * knot_count


@jax.jit(static_argnames=("rollout_count", "knot_count"))
@jaxtyped
def sample_halton(
    *,
    around: Float[JaxArray, "T D_u"],
    standard_deviation: Float[JaxArray, " D_u"],
    rollout_count: int,
    knot_count: int,
    halton_start_index: IntScalar,
) -> tuple[Float[JaxArray, "T D_u M"], IntScalar]:
    time_horizon, control_dimension = around.shape
    halton_dimensions = halton_dimensions_for(
        control_dimension=control_dimension, knot_count=knot_count
    )

    primes = PRIMES[:halton_dimensions]
    uniform_samples = generate_halton_sequence(
        primes, rollout_count, halton_start_index
    )

    gaussian_knots = transform_halton_to_knots(
        uniform_samples, knot_count, control_dimension
    )

    knot_times = jnp.linspace(0, time_horizon - 1, knot_count)
    evaluation_times = jnp.arange(time_horizon, dtype=jnp.float32)

    perturbations = interpolate_knots(gaussian_knots, knot_times, evaluation_times)
    scaled_perturbations = perturbations * standard_deviation[None, :, None]

    samples = around[..., None] + scaled_perturbations
    new_halton_index = halton_start_index + rollout_count

    return samples, new_halton_index


@jax.jit(static_argnames=("count",))
@jaxtyped
def generate_halton_sequence(
    primes: Int[JaxArray, " D"], count: int, start_index: IntScalar
) -> Float[JaxArray, "count D"]:
    indices = jnp.arange(count) + start_index

    def compute_sample(index: Int[JaxArray, ""]) -> Float[JaxArray, " D"]:
        return jax.vmap(lambda base: radical_inverse(index, base))(primes)

    return jax.vmap(compute_sample)(indices)


@jax.jit(static_argnames=("knot_count", "control_dimension"))
@jaxtyped
def transform_halton_to_knots(
    uniform_samples: Float[JaxArray, "M halton_dim"],
    knot_count: int,
    control_dimension: int,
) -> Float[JaxArray, "M K D_u"]:
    gaussian_samples = inverse_normal_cdf(uniform_samples)
    return gaussian_samples.reshape(-1, knot_count, control_dimension)


@jax.jit
@jaxtyped
def radical_inverse(
    index: Int[JaxArray, ""], base: Int[JaxArray, ""]
) -> Float[JaxArray, ""]:
    def body_fn(
        carry: RadicalInverseCarry, _: None
    ) -> tuple[RadicalInverseCarry, None]:
        current_index, result, factor = carry
        digit = current_index % base
        result = result + digit * factor
        current_index = current_index // base
        factor = factor / base
        return (current_index, result, factor), None

    initial_state: RadicalInverseCarry = (index, jnp.array(0.0), 1.0 / base)
    (_, result, _), _ = jax.lax.scan(
        body_fn, initial_state, None, length=MAX_RADICAL_INVERSE_ITERATIONS
    )

    return result


@jax.jit
@jaxtyped
def inverse_normal_cdf(
    uniform_values: Float[JaxArray, "*batch"],
) -> Float[JaxArray, "*batch"]:
    clamped = jnp.clip(uniform_values, 1e-7, 1 - 1e-7)
    return jnp.sqrt(2.0) * jspecial.erfinv(2.0 * clamped - 1.0)


@jax.jit
@jaxtyped
def solve_tridiagonal(
    lower: Float[JaxArray, " N"],
    diag: Float[JaxArray, " N"],
    upper: Float[JaxArray, " N"],
    rhs: Float[JaxArray, " N"],
) -> Float[JaxArray, " N"]:
    # NOTE: Thomas algorithm can be used, but the knot counts are small anyways.
    matrix = jnp.diag(diag) + jnp.diag(upper[:-1], k=1) + jnp.diag(lower[1:], k=-1)
    return jnp.linalg.solve(matrix, rhs)


@jax.jit
@jaxtyped
def compute_spline_coefficients(
    knot_times: Float[JaxArray, " K"],
    knot_values: Float[JaxArray, " K"],
) -> Float[JaxArray, " K"]:
    h = jnp.diff(knot_times)
    b = jnp.diff(knot_values) / h

    diag = jnp.concatenate([jnp.array([1.0]), 2 * (h[:-1] + h[1:]), jnp.array([1.0])])
    upper = jnp.concatenate([jnp.array([0.0]), h[1:], jnp.array([0.0])])
    lower = jnp.concatenate([jnp.array([0.0]), h[:-1], jnp.array([0.0])])
    rhs = jnp.concatenate([jnp.array([0.0]), 3 * (b[1:] - b[:-1]), jnp.array([0.0])])

    return solve_tridiagonal(lower, diag, upper, rhs)


@jax.jit
@jaxtyped
def evaluate_cubic_spline(
    knot_times: Float[JaxArray, " K"],
    knot_values: Float[JaxArray, " K"],
    second_derivatives: Float[JaxArray, " K"],
    evaluation_times: Float[JaxArray, " T"],
) -> Float[JaxArray, " T"]:
    knot_count = knot_times.shape[0]

    def evaluate_at_time(t: Float[JaxArray, ""]) -> Float[JaxArray, ""]:
        i = jnp.clip(
            jnp.searchsorted(knot_times, t, side="right") - 1, 0, knot_count - 2
        )

        t0, t1 = knot_times[i], knot_times[i + 1]
        y0, y1 = knot_values[i], knot_values[i + 1]
        d0, d1 = second_derivatives[i], second_derivatives[i + 1]

        h = t1 - t0
        a, b = t1 - t, t - t0

        return (
            (a / h) * y0
            + (b / h) * y1
            + ((a**3 / h - a * h) * d0 + (b**3 / h - b * h) * d1) / 6.0
        )

    return jax.vmap(evaluate_at_time)(evaluation_times)


@jax.jit
@jaxtyped
def interpolate_single_spline(
    knot_times: Float[JaxArray, " K"],
    knot_values: Float[JaxArray, " K"],
    evaluation_times: Float[JaxArray, " T"],
) -> Float[JaxArray, " T"]:
    second_derivatives = compute_spline_coefficients(knot_times, knot_values)
    return evaluate_cubic_spline(
        knot_times, knot_values, second_derivatives, evaluation_times
    )


@jax.jit
@jaxtyped
def interpolate_knots(
    gaussian_knots: Float[JaxArray, "M K D_u"],
    knot_times: Float[JaxArray, " K"],
    evaluation_times: Float[JaxArray, " T"],
) -> Float[JaxArray, "T D_u M"]:
    def interpolate_rollout(
        rollout_knots: Float[JaxArray, "K D_u"],
    ) -> Float[JaxArray, "T D_u"]:
        return jax.vmap(
            lambda knots: interpolate_single_spline(
                knot_times, knots, evaluation_times
            ),
            in_axes=1,
            out_axes=1,
        )(rollout_knots)

    perturbations = jax.vmap(interpolate_rollout)(gaussian_knots)
    return jnp.transpose(perturbations, (1, 2, 0))



from .kf import (
    NumPyKalmanFilter as NumPyKalmanFilter,
    numpy_kalman_filter as numpy_kalman_filter,
    JaxKalmanFilter as JaxKalmanFilter,
    jax_kalman_filter as jax_kalman_filter,
)
from .ekf import (
    NumPyExtendedKalmanFilter as NumPyExtendedKalmanFilter,
    JaxExtendedKalmanFilter as JaxExtendedKalmanFilter,
)
from .ukf import (
    NumPyUnscentedKalmanFilter as NumPyUnscentedKalmanFilter,
    JaxUnscentedKalmanFilter as JaxUnscentedKalmanFilter,
)
from .noise import (
    NumPyAdaptiveNoise as NumPyAdaptiveNoise,
    NumPyAdaptiveNoiseProvider as NumPyAdaptiveNoiseProvider,
    NumPyClampedNoise as NumPyClampedNoise,
    NumPyClampedNoiseProvider as NumPyClampedNoiseProvider,
    NumPyNoiseCovarianceBounds as NumPyNoiseCovarianceBounds,
    JaxAdaptiveNoise as JaxAdaptiveNoise,
    JaxAdaptiveNoiseProvider as JaxAdaptiveNoiseProvider,
    JaxClampedNoise as JaxClampedNoise,
    JaxClampedNoiseProvider as JaxClampedNoiseProvider,
    JaxNoiseCovarianceBounds as JaxNoiseCovarianceBounds,
    IdentityNoiseModelProvider as IdentityNoiseModelProvider,
)
from .factory import noise as noise



from .common import (
    EstimatedObstacleStates as EstimatedObstacleStates,
    ObstacleStatesHistory as ObstacleStatesHistory,
    ObstacleIds as ObstacleIds,
    ObstacleIdAssignment as ObstacleIdAssignment,
    ObstacleStatesRunningHistory as ObstacleStatesRunningHistory,
    ObstacleModel as ObstacleModel,
    ObstacleStateEstimator as ObstacleStateEstimator,
    InputAssumptionProvider as InputAssumptionProvider,
    PredictionCreator as PredictionCreator,
    ObstacleMotionPredictor as ObstacleMotionPredictor,
)
from .basic import (
    NumPyObstacleStatesHistory as NumPyObstacleStatesHistory,
)
from .accelerated import (
    JaxObstacleStatesHistory as JaxObstacleStatesHistory,
)



from .basic import (
    NumPyGaussianObstacle2dPoseSampler as NumPyGaussianObstacle2dPoseSampler,
)
from .accelerated import (
    JaxGaussianObstacle2dPoseSampler as JaxGaussianObstacle2dPoseSampler,
)



from typing import Any, Protocol

from faran.types.array import ArrayConvertible


class Noise(Protocol):
    @property
    def process_noise_covariance(self) -> ArrayConvertible:
        """Returns the process noise covariance matrix."""
        ...

    @property
    def observation_noise_covariance(self) -> ArrayConvertible:
        """Returns the observation noise covariance matrix."""
        ...


class NoiseModel[NoiseT, BeliefT, ObservationT, StateT = Any](Protocol):
    def __call__(
        self,
        *,
        noise: NoiseT,
        prediction: BeliefT,
        observation: ObservationT,
        state: StateT,
    ) -> tuple[NoiseT, StateT]:
        """Returns the updated noise covariances and updated state."""
        ...

    @property
    def state(self) -> StateT:
        """Returns the initial state for this noise model."""
        ...


class NoiseModelProvider[NoiseT, BeliefT, ObservationT, MatrixT, StateT = Any](
    Protocol
):
    def __call__(
        self, *, observation_matrix: MatrixT, noise: NoiseT
    ) -> NoiseModel[NoiseT, BeliefT, ObservationT, StateT]:
        """Creates a noise model using the specified observation matrix and
        initial noise covariances."""
        ...



from typing import Final
from dataclasses import dataclass

from faran.types import (
    Array,
    NumPyControlInputBatchCreator,
    NumPyControlInputSequence,
    NumPyControlInputBatch,
    NumPySampler,
)
from faran.samplers.basic import StandardDeviationDescription, standardize

from jaxtyping import Float

import numpy as np


@dataclass(frozen=True)
class NumPyGaussianSampler[
    BatchT: NumPyControlInputBatch,
](NumPySampler[NumPyControlInputSequence, BatchT]):
    """Perturbs a nominal control sequence with zero-mean Gaussian noise."""

    standard_deviation: Final[Float[Array, " D_u"]]
    to_batch: Final[NumPyControlInputBatchCreator[BatchT]]
    rng: np.random.Generator

    _rollout_count: Final[int]

    @staticmethod
    def create[B: NumPyControlInputBatch](
        *,
        standard_deviation: StandardDeviationDescription,
        rollout_count: int,
        to_batch: NumPyControlInputBatchCreator,
        seed: int,
    ) -> "NumPyGaussianSampler":
        """Creates a sampler generating Gaussian noise around the specified control input
        sequence.
        """
        return NumPyGaussianSampler(
            standard_deviation=standardize.std(standard_deviation),
            to_batch=to_batch,
            rng=np.random.default_rng(seed),
            _rollout_count=rollout_count,
        )

    def sample(self, *, around: NumPyControlInputSequence) -> BatchT:
        samples = around.array[..., None] + self.rng.normal(
            loc=0.0,
            scale=self.standard_deviation[None, :, None],
            size=(around.horizon, around.dimension, self.rollout_count),
        )

        return self.to_batch(array=samples)

    @property
    def rollout_count(self) -> int:
        return self._rollout_count



from typing import Protocol
from dataclasses import dataclass

from faran.types.array import Array, DataType, jaxtyped
from faran.types.costs.boundary.common import BoundaryDistanceExtractor

from jaxtyping import Float


class NumPyBoundaryDistanceExtractor[StateBatchT, DistanceT](
    BoundaryDistanceExtractor[StateBatchT, DistanceT], Protocol
): ...


@jaxtyped
@dataclass(frozen=True)
class NumPyBoundaryDistance:
    _array: Float[Array, "T M"]

    @staticmethod
    def create(*, array: Float[Array, "T M"]) -> "NumPyBoundaryDistance":
        """Creates a NumPy boundary distance from the given array."""
        return NumPyBoundaryDistance(array)

    def __array__(self, dtype: DataType | None = None) -> Float[Array, "T M"]:
        return self._array

    @property
    def horizon(self) -> int:
        return self._array.shape[0]

    @property
    def rollout_count(self) -> int:
        return self._array.shape[1]

    @property
    def array(self) -> Float[Array, "T M"]:
        return self._array



from typing import Any
from dataclasses import dataclass

from faran.types import (
    ObstacleStatesForTimeStep,
    ObstacleStateProvider,
    ObstacleMotionPredictor,
    ObstacleStatesRunningHistory,
    ObstacleIdAssignment,
)


class DefaultIdAssignment(ObstacleIdAssignment[Any, Any, Any]):
    """No-op obstacle ID assignment that preserves existing IDs."""

    def __call__(self, states: Any, /, *, history: Any, ids: Any) -> None:
        return None


@dataclass(kw_only=True)
class PredictingObstacleStateProvider[
    ObstacleStatesForTimeStepT,
    IdT,
    HistoryT,
    PredictionT,
](ObstacleStateProvider[PredictionT]):
    """Provides obstacle state predictions by maintaining a running history and forwarding to a predictor."""

    predictor: ObstacleMotionPredictor[HistoryT, PredictionT]
    history: ObstacleStatesRunningHistory[ObstacleStatesForTimeStepT, IdT, HistoryT]
    id_assignment: ObstacleIdAssignment[ObstacleStatesForTimeStepT, IdT, HistoryT]

    @staticmethod
    def create[O: ObstacleStatesForTimeStep, I, H, P](
        *,
        predictor: ObstacleMotionPredictor[H, P],
        history: ObstacleStatesRunningHistory[O, I, H],
        id_assignment: ObstacleIdAssignment[O, I, H] | None = None,
    ) -> "PredictingObstacleStateProvider[O, I, H, P]":
        return PredictingObstacleStateProvider(
            predictor=predictor,
            history=history,
            id_assignment=id_assignment
            if id_assignment is not None
            else DefaultIdAssignment(),
        )

    def __call__(self) -> PredictionT:
        return self.predictor.predict(history=self.history.get())

    def observe(self, states: ObstacleStatesForTimeStepT) -> None:
        self.history = self.history.append(
            states,
            ids=self.id_assignment(
                states, history=self.history.get(), ids=self.history.ids()
            ),
        )



from dataclasses import dataclass

from faran.types.array import Array, jaxtyped

from jaxtyping import Float

type OriginsArray = Float[Array, "V 2"]
type RadiiArray = Float[Array, " V"]


@jaxtyped
@dataclass(frozen=True)
class Circles:
    """Describes circles approximating parts of an object for distance computation."""

    origins: OriginsArray
    """The local (x, y) offset of the circle center from the object center."""

    radii: RadiiArray
    """The radius of the circle."""
